import functools
import logging
import shlex
from itertools import starmap
from subprocess import check_output
from typing import List, Optional

from capstone import CsInsn
from strongarm_dataflow.dataflow import get_register_contents_at_instruction_fast
from strongarm_dataflow.register_contents import RegisterContents, RegisterContentsType

from strongarm.macho import MachoBinary, ObjcClass, ObjcSelector, VirtualMemoryPointer

from .objc_instruction import ObjcBranchInstruction, ObjcInstruction, ObjcUnconditionalBranchInstruction


def _is_mangled_cpp_symbol(symbol_name: str) -> bool:
    """Return whether a symbol name appears to be a mangled C++ symbol.
    """
    return any(symbol_name.startswith(prefix) for prefix in ["_Z", "__Z", "___Z"])


def _demangle_cpp_symbol(cpp_symbol: str) -> str:
    """Call into c++filt to demangle the provided mangled C++ symbol name.
    """
    if not _is_mangled_cpp_symbol(cpp_symbol):
        return cpp_symbol

    original_symbol = cpp_symbol

    # Linux's c++filt doesn't like the clang-specific "_block_invoke" which is tacked onto ObjC++ blocks.
    # Trim this off and add it back after demangling the symbol
    is_block = False
    block_index = ""
    if "_block_invoke" in cpp_symbol:
        is_block = True
        cpp_symbol, block_index_str = cpp_symbol.split("_block_invoke")
        # Some blocks have an index
        if block_index_str.isnumeric():
            block_index = f" {int(block_index_str)}"

    # XXX(PT): We observe that c++filt doesn't work if there are too many leading underscores
    # Try demangling multiple times, trimming a leading underscore each time until success (up to 3 times)
    for _ in range(3):
        # If demangling fails, allow the exception to propagate up. This can alert us to scanner issues.
        demangled_symbol = check_output(["c++filt", "-_", shlex.quote(cpp_symbol)]).decode().strip()
        # Was the symbol demangled?
        if demangled_symbol != cpp_symbol:
            if is_block:
                demangled_symbol = f"block{block_index} in {demangled_symbol}"
            return demangled_symbol
        else:
            # Trim an underscore and try again if possible
            if not cpp_symbol.startswith("_"):
                break
            cpp_symbol = cpp_symbol[1:]

    # Failed to demangle, return the original symbol name
    return original_symbol


class ObjcMethodInfo:

    __slots__ = ["objc_class", "objc_sel", "imp_addr"]

    def __init__(self, objc_class: ObjcClass, objc_sel: ObjcSelector, imp: Optional[VirtualMemoryPointer]) -> None:
        self.objc_class = objc_class
        self.objc_sel = objc_sel
        self.imp_addr = imp

    def __repr__(self) -> str:
        return f"-[{self.objc_class.name} {self.objc_sel.name}]"


class BasicBlock:
    def __init__(self, start_address: VirtualMemoryPointer, end_address: VirtualMemoryPointer) -> None:
        """Represents a basic-block of assembly code.

        A 'basic block' is a unit of assembly code with no branching except for the last instruction.
        In other words, it is the smallest unit of callable code - a subroutine.
        There is a single entry point, and single exit point.

        The start and end addresses are inclusive.
        """
        self.start_address = start_address
        self.end_address = end_address


class ObjcFunctionAnalyzer:
    """Provides utility functions for introspecting on a set of instructions which represent a function body.
    As Objective-C is a strict superset of C, ObjcFunctionAnalyzer can also be used on pure C functions.
    """

    def __init__(self, binary: MachoBinary, instructions: List[CsInsn], method_info: ObjcMethodInfo = None) -> None:
        from strongarm.macho import MachoAnalyzer

        try:
            self.start_address = VirtualMemoryPointer(instructions[0].address)
            last_instruction = instructions[len(instructions) - 1]
            # The end-address is right-exclusive
            self.end_address = VirtualMemoryPointer(last_instruction.address) + MachoBinary.BYTES_PER_INSTRUCTION
        except IndexError:
            # this method must have just been a stub with no real instructions!
            self.start_address = VirtualMemoryPointer(0)
            self.end_address = VirtualMemoryPointer(0)

        self.binary = binary
        self.macho_analyzer = MachoAnalyzer.get_analyzer(binary)
        self.instructions = instructions
        self.method_info = method_info

        self._call_targets: Optional[List[ObjcBranchInstruction]] = None

        # Find basic-block-boundaries upfront
        self.basic_blocks = self._find_basic_blocks()

    def _get_instruction_index_of_address(self, address: VirtualMemoryPointer) -> Optional[int]:
        """Return the index of an instruction with a provided address within the internal list of instructions
        """
        base_address = self.start_address
        offset = address - base_address
        index = int(offset / MachoBinary.BYTES_PER_INSTRUCTION)
        if 0 <= index < len(self.instructions):
            return index
        return None

    def get_instruction_at_index(self, index: int) -> Optional[CsInsn]:
        """Get the instruction at a given index within the function's code, wrapping in ObjcInstruction
        """
        if 0 <= index < len(self.instructions):
            return self.instructions[index]
        return None

    def get_instruction_at_address(self, address: VirtualMemoryPointer) -> Optional[CsInsn]:
        """Get the Instruction within the analyzed function at a provided address.
        The return will be wrapped in an ObjcInstruction.
        This method will return None if the address is not contained within the analyzed function.
        """
        index = self._get_instruction_index_of_address(address)
        if index is None:
            return None
        return self.get_instruction_at_index(index)

    def debug_print(self, idx: int, output: str) -> None:
        """Helper function to pretty-print debug logs

        Args:
            idx: instruction offset within function the message references
            output: string to output to debug log
        """
        if not len(self.instructions):
            logging.debug(self, f"func(stub) {output}")
        else:
            func_base = self.start_address
            instruction_address = func_base + (idx * MachoBinary.BYTES_PER_INSTRUCTION)
            logging.debug(self, f"func({hex(int(instruction_address))}) {output}")

    def get_symbol_name(self) -> str:
        """Return a objective-c class/method, c function, or sub_address-style string representing the name of
            this block of code.
        """
        if self.method_info:
            return f"-[{self.method_info.objc_class.name} {self.method_info.objc_sel.name}]"
        else:
            # Not an Objective-C method. Try to find a symbol name that matches the address
            strtbl_sym_name = self.macho_analyzer.crossref_helper.get_symbol_name_for_address(
                VirtualMemoryPointer(self.start_address)
            )

            if strtbl_sym_name:
                # Demangle C++ symbols when applicable
                if _is_mangled_cpp_symbol(strtbl_sym_name):
                    strtbl_sym_name = _demangle_cpp_symbol(strtbl_sym_name)

                return strtbl_sym_name

        # Fallback
        # We don't want to format the procedure as sub_<address>, because we use the output of this method to
        # report code locations, and the address of the same procedure might change between subsequent binary builds.
        return "_unsymbolicated_function"

    @classmethod
    def get_function_analyzer(cls, binary: MachoBinary, start_address: VirtualMemoryPointer) -> "ObjcFunctionAnalyzer":
        """Get the shared analyzer for the function at start_address in the binary.

        This method uses a cached MachoAnalyzer if available, which is more efficient than analyzing the
        same binary over and over. Therefore, this method should be used when an ObjcFunctionAnalyzer is needed,
        instead of constructing it yourself.

        Args:
            binary: The MachoBinary containing a function at start_address
            start_address: The entry point address for the function to be analyzed

        Returns:
            An ObjcFunctionAnalyzer suitable for introspecting a block of code.
        """
        from strongarm.macho.macho_analyzer import MachoAnalyzer

        analyzer = MachoAnalyzer.get_analyzer(binary)
        instructions = analyzer.get_function_instructions(start_address)
        return ObjcFunctionAnalyzer(binary, instructions)

    @classmethod
    def get_function_analyzer_for_method(
        cls, binary: MachoBinary, method_info: ObjcMethodInfo
    ) -> "ObjcFunctionAnalyzer":
        """Get the shared analyzer describing an Objective-C method within the Mach-O binary
        This method performs the same caching as get_function_analyzer()

        Args:
            binary: The MachoBinary containing a function at method_info.imp_addr
            method_info: The ObjcMethodInfo describing the IMP to be analyzed

        Returns:
            An ObjcFunctionAnalyzer suitable for introspecting the provided method

        Raises:
            ValueError: Could not get function instructions for the provided method
        """
        # TODO(PT): it seems like this & related methods should be moved to MachoAnalyzer
        if not method_info.imp_addr:
            raise ValueError(f"Could not get method implementation address for {method_info}")

        from strongarm.macho.macho_analyzer import MachoAnalyzer

        analyzer = MachoAnalyzer.get_analyzer(binary)
        instructions = analyzer.get_function_instructions(method_info.imp_addr)
        return ObjcFunctionAnalyzer(binary, instructions, method_info=method_info)

    @classmethod
    def get_function_analyzer_for_signature(
        cls, binary: MachoBinary, class_name: str, sel_name: str
    ) -> "ObjcFunctionAnalyzer":
        from strongarm.macho.macho_analyzer import MachoAnalyzer

        analyzer = MachoAnalyzer.get_analyzer(binary)
        for objc_cls in analyzer.objc_classes():
            if objc_cls.name == class_name:
                for sel in objc_cls.selectors:
                    if sel.name == sel_name:
                        # XXX(PT): where are the method info's normally stored? Can we grab it from there?
                        method_info = ObjcMethodInfo(objc_cls, sel, sel.implementation)
                        return ObjcFunctionAnalyzer.get_function_analyzer_for_method(binary, method_info)
        raise RuntimeError(f"No found function analyzer for -[{class_name} {sel_name}]")

    @property
    def call_targets(self) -> List[ObjcBranchInstruction]:
        """Return the List of all branch instructions within the source function.
        """
        # Use cached list if available
        if self._call_targets is not None:
            return self._call_targets

        # Extract the list of branch instructions in the function
        branches_in_function: List[ObjcBranchInstruction] = []
        for idx, instr in enumerate(self.instructions):
            if ObjcBranchInstruction.is_branch_instruction(instr):
                branches_in_function.append(
                    ObjcBranchInstruction.parse_instruction(
                        self, instr, container_function_boundary=(self.start_address, self.end_address)
                    )
                )

        self._call_targets = branches_in_function
        return self._call_targets

    @property
    def function_call_targets(self) -> List["ObjcFunctionAnalyzer"]:
        """Find List of function analyzers representing functions reachable from the source function.

        This excludes other branch destinations, such as objc_msgSend calls to methods implemented outside this
        binary, or local branching within the source function.
        """
        call_targets = []
        for target in self.call_targets:
            # don't try to follow calls to functions defined outside this binary
            if target.is_external_c_call and not target.is_msgSend_call:
                continue
            # don't try to follow path if it's an internal branch (i.e. control flow within this function)
            # any internal branching will eventually be covered by call_targets,
            # so there's no need to follow twice
            if self.is_local_branch(target):
                continue
            # might be objc_msgSend to object of class defined outside binary
            if target.is_external_objc_call:
                continue
            call_targets.append(ObjcFunctionAnalyzer.get_function_analyzer(self.binary, target.destination_address))
        return call_targets

    def get_local_branches(self) -> List[ObjcBranchInstruction]:
        """Return all instructions in the analyzed function representing a branch to a destination within the function
        """
        local_branches = []
        for target in self.call_targets:
            # find the address of this branch instruction within the function
            if self.is_local_branch(target):
                local_branches.append(target)
        return local_branches

    @classmethod
    def format_instruction(cls, instr: CsInsn) -> str:
        """Stringify a CsInsn for printing
        Args:
            instr: Instruction to create formatted string representation for
        Returns:
            Formatted string representing instruction
        """
        return f"{hex(int(instr.address))}:\t{instr.mnemonic}\t{instr.op_str}"

    def is_local_branch(self, branch_instruction: ObjcBranchInstruction) -> bool:
        # if there's no destination address, the destination is outside the binary, and it couldn't possible be local
        if not branch_instruction.destination_address:
            return False
        return self.start_address <= branch_instruction.destination_address <= self.end_address

    def get_objc_selref(self, msgsend_instr: ObjcUnconditionalBranchInstruction) -> VirtualMemoryPointer:
        """Returns the selref pointer at an _objc_msgSend call site.
        When _objc_msgSend is called, x1 contains the selref being messaged.
        The caller is responsible for ensuring this is called at an _objc_msgSend call site.
        """
        if msgsend_instr.raw_instr.mnemonic not in ObjcUnconditionalBranchInstruction.UNCONDITIONAL_BRANCH_MNEMONICS:
            raise ValueError("get_objc_selref() called on non-branch instruction")

        # at an _objc_msgSend call site, the selref is in x1
        contents = self.get_register_contents_at_instruction("x1", msgsend_instr)
        if contents.type != RegisterContentsType.IMMEDIATE:
            raise RuntimeError(f"could not determine selref ptr, origates in function arg (type {contents.type.name})")
        return VirtualMemoryPointer(contents.value)

    @functools.lru_cache(maxsize=100)
    def get_register_contents_at_instruction(self, register: str, instruction: ObjcInstruction) -> RegisterContents:
        # If basic-block analysis has been done, reduce the dataflow analysis space to the instruction's basic-block
        # Otherwise, use the entire source function as the search space
        for bb in self.basic_blocks:
            if bb.start_address <= instruction.address < bb.end_address:
                # Found the basic block containing the instruction; reduce dataflow analysis space to its head
                dataflow_space_start = bb.start_address
                dataflow_space_end = bb.end_address
                break
        else:
            # We are in the process of computing basic blocks, so we can't query them. Use the whole function for DFA
            dataflow_space_start = self.start_address
            dataflow_space_end = self.end_address

        # To try and save a bit of work, don't include bytecode past the end of this basic block,
        # as we only need the bytecode up to the provided instruction
        function_bytecode = self.binary.get_content_from_virtual_address(
            self.start_address, dataflow_space_end - self.start_address
        )
        return get_register_contents_at_instruction_fast(
            register, self.start_address, function_bytecode, dataflow_space_start, instruction.address
        )

    def _find_basic_blocks(self) -> List["BasicBlock"]:
        """Locate the basic-block-boundaries within the source function.
        A 'basic block' is a unit of assembly code with no branching except for the last instruction.
        In other words, it is the smallest unit of callable code - a subroutine.
        There is a single entry point, and single exit point.

        Returns:
            a List of objects encapsulating the basic block boundaries.
        """
        basic_blocks = self.macho_analyzer.get_basic_block_boundaries(self.start_address)
        return list(starmap(BasicBlock, basic_blocks))

    def __repr__(self) -> str:
        return f"({self.get_symbol_name()} @ {self.start_address})"



import logging
from ctypes import c_uint32, c_uint64, sizeof
from typing import Dict, List, Optional

from more_itertools import first_true

from strongarm.macho.arch_independent_structs import (
    ArchIndependentStructure,
    ObjcCategoryRawStruct,
    ObjcClassRawStruct,
    ObjcDataRawStruct,
    ObjcIvarListStruct,
    ObjcIvarStruct,
    ObjcMethodListStruct,
    ObjcMethodStruct,
    ObjcProtocolListStruct,
    ObjcProtocolRawStruct,
)
from strongarm.macho.macho_binary import MachoBinary
from strongarm.macho.macho_definitions import VirtualMemoryPointer


class ObjcSelref:
    __slots__ = ["source_address", "destination_address", "selector_literal"]

    def __init__(
        self, source_address: VirtualMemoryPointer, destination_address: VirtualMemoryPointer, selector_literal: str
    ) -> None:
        self.source_address = source_address
        self.destination_address = destination_address
        self.selector_literal = selector_literal

    def __repr__(self) -> str:
        return (
            f"<ObjcSelref source=0x{self.source_address:x} dest=0x{self.destination_address:x}"
            f" sel={self.selector_literal}>"
        )


class ObjcSelector:
    __slots__ = ["name", "selref", "implementation", "is_external_definition"]

    def __init__(self, name: str, selref: Optional[ObjcSelref], implementation: Optional[VirtualMemoryPointer]) -> None:
        self.name = name
        self.selref = selref
        self.implementation = implementation

        self.is_external_definition = not self.implementation

    def __str__(self) -> str:
        imp_addr = "NaN"
        if self.implementation:
            imp_addr = hex(int(self.implementation))
        return f"<@selector({self.name}) at {imp_addr}>"

    __repr__ = __str__


class ObjcIvar:
    __slots__ = ["name", "class_name", "field_offset", "field_offset_addr"]

    def __init__(self, name: str, class_name: str, offset: int, field_offset_addr: VirtualMemoryPointer):
        self.name = name
        self.class_name = class_name
        self.field_offset = offset
        self.field_offset_addr = field_offset_addr

    def __str__(self) -> str:
        return f"<@ivar {self.class_name}* {self.name}, off @ {self.field_offset}>"

    __repr__ = __str__


class ObjcClass:
    __slots__ = ["raw_struct", "name", "selectors", "ivars", "protocols", "super_classref", "superclass_name"]

    def __init__(
        self,
        raw_struct: ArchIndependentStructure,
        name: str,
        selectors: List[ObjcSelector],
        ivars: List[ObjcIvar] = None,
        protocols: List["ObjcProtocol"] = None,
        super_classref: Optional[VirtualMemoryPointer] = None,
        superclass_name: Optional[str] = None,
    ) -> None:
        self.name = name
        self.selectors = selectors
        self.raw_struct = raw_struct
        self.ivars = ivars if ivars else []
        self.protocols = protocols if protocols else []
        self.super_classref = super_classref
        self.superclass_name = superclass_name

    def __str__(self) -> str:
        return f"ObjcClass({self.name} : {self.superclass_name})"

    def __repr__(self) -> str:
        return (
            f"<@class {self.name} : {self.superclass_name}"
            f" sel_count={len(self.selectors)} ivar_count={len(self.ivars)} protocol_count={len(self.protocols)}>"
        )


class ObjcProtocol(ObjcClass):
    pass


class ObjcCategory(ObjcClass):
    __slots__ = ["raw_struct", "name", "base_class", "category_name", "selectors", "ivars", "protocols"]

    def __init__(
        self,
        raw_struct: ObjcCategoryRawStruct,
        base_class: str,
        category_name: str,
        selectors: List[ObjcSelector],
        ivars: List[ObjcIvar] = None,
        protocols: List[ObjcProtocol] = None,
    ) -> None:
        self.base_class = base_class
        self.category_name = category_name

        # ObjcCategory.name includes the base class + the cat-name
        # That way, callers don't need to check the ObjcClass instance type to get the 'right' value
        full_name = f"{base_class} ({category_name})"
        super().__init__(raw_struct, full_name, selectors, ivars, protocols)

    def __str__(self) -> str:
        return f"ObjcCategory({self.base_class} ({self.category_name}))"

    def __repr__(self) -> str:
        return (
            f"<@class {self.base_class} ({self.category_name})"
            f" sel_count={len(self.selectors)} ivar_count={len(self.ivars)} protocol_count={len(self.protocols)}>"
        )


class ObjcRuntimeDataParser:
    def __init__(self, binary: MachoBinary) -> None:
        self.binary = binary
        logging.debug(f"Parsing ObjC runtime info of {self.binary}...")

        logging.debug("Step 1: Parsing selrefs...")
        self._selref_ptr_to_selector_map: Dict[VirtualMemoryPointer, ObjcSelector] = {}
        self._selector_literal_ptr_to_selref_map: Dict[VirtualMemoryPointer, ObjcSelref] = {}
        # This populates self._selector_literal_ptr_to_selref_map and self._selref_ptr_to_selector_map
        self._parse_selrefs()

        logging.debug("Step 2: Parsing classes, categories, and protocols...")
        self._classrefs_to_objc_classes: Dict[VirtualMemoryPointer, ObjcClass] = {}
        # This populates self._classrefs_to_objc_classes
        self.classes = self._parse_class_and_category_info()
        self.protocols = self._parse_global_protocol_info()

        logging.debug("Step 3: Resolving symbol name to source dylib map...")
        self._sym_to_dylib_path = self._parse_linked_dylib_symbols()

    def _parse_linked_dylib_symbols(self) -> Dict[str, str]:
        syms_to_dylib_path = {}

        symtab = self.binary.symtab
        symtab_contents = self.binary.symtab_contents
        dysymtab = self.binary.dysymtab
        visited_addresses = set()
        for undef_sym_idx in range(dysymtab.nundefsym):
            symtab_idx = dysymtab.iundefsym + undef_sym_idx
            sym = symtab_contents[symtab_idx]

            strtab_idx = sym.n_un.n_strx
            string_file_address = symtab.stroff + strtab_idx

            # Some binaries contain a symtab such that all the calculated string address are the same. This check
            # prevents spamming the logs with errors about the same symbol
            # TODO(FS): Task tracking this issue SCAN-2744
            if string_file_address in visited_addresses:
                continue
            visited_addresses.add(string_file_address)

            symbol_name = self.binary.get_full_string_from_start_address(string_file_address, virtual=False)
            if not symbol_name:
                logging.error(f"Could not get symbol name at address {hex(string_file_address)}")
                continue

            library_ordinal = self._library_ordinal_from_n_desc(sym.n_desc)
            source_name = self.binary.dylib_name_for_library_ordinal(library_ordinal)

            syms_to_dylib_path[symbol_name] = source_name
        return syms_to_dylib_path

    def path_for_external_symbol(self, symbol: str) -> Optional[str]:
        if symbol in self._sym_to_dylib_path:
            return self._sym_to_dylib_path[symbol]
        return None

    @staticmethod
    def _library_ordinal_from_n_desc(n_desc: int) -> int:
        return (n_desc >> 8) & 0xFF

    def _parse_selrefs(self) -> None:
        """Parse the binary's selref list, and store the data.

        This method populates self._selector_literal_ptr_to_selref_map.
        It also *PARTLY* populates self._selref_ptr_to_selector_map. All selrefs keys will have an ObjcSelector
        value, but none of the ObjcSelector objects will have their `implementation` field filled, because
        at this point in the parse we do not yet know the implementations of each selector. ObjcSelectors which we
        later find an implementation for are updated in self.read_selectors_from_methlist_ptr"""
        selref_pointers, selector_literal_pointers = self.binary.read_pointer_section("__objc_selrefs")
        # sanity check
        if len(selref_pointers) != len(selector_literal_pointers):
            raise RuntimeError("read invalid data from __objc_selrefs")

        for selref_ptr, selector_literal_ptr in zip(selref_pointers, selector_literal_pointers):
            # read selector string literal from selref pointer
            selector_string = self.binary.get_full_string_from_start_address(selector_literal_ptr)
            if not selector_string:
                continue  # but all selectors should have a name
            wrapped_selref = ObjcSelref(selref_ptr, selector_literal_ptr, selector_string)

            # map the selector string pointer to the ObjcSelref
            self._selector_literal_ptr_to_selref_map[selector_literal_ptr] = wrapped_selref
            # add second mapping in selref list
            # we don't know the implementation address yet but it will be updated when we parse method lists
            self._selref_ptr_to_selector_map[selref_ptr] = ObjcSelector(selector_string, wrapped_selref, None)

    def selector_for_selref(self, selref_addr: VirtualMemoryPointer) -> Optional[ObjcSelector]:
        selector = self._selref_ptr_to_selector_map.get(selref_addr)
        if selector is not None:
            return selector

        # selref wasn't referenced in classes implemented within the binary
        # make sure it's a valid selref
        selrefs = self._selector_literal_ptr_to_selref_map.values()
        selref = first_true(iter(selrefs), pred=lambda x: x.source_address == selref_addr, default=None)

        if selref is not None:
            # Therefore, the selref must refer to a selector which is defined outside this binary
            # this is fine, just construct an ObjcSelector with what we know
            return ObjcSelector(selref.selector_literal, selref, None)

        else:
            return None

    def selector_for_selector_literal(self, literal_addr: VirtualMemoryPointer) -> Optional[ObjcSelector]:
        selector_literal = self._selector_literal_ptr_to_selref_map.get(literal_addr)
        if selector_literal is not None:
            return self.selector_for_selref(selector_literal.source_address)
        else:
            return None

    def selrefs_to_selectors(self) -> Dict[VirtualMemoryPointer, ObjcSelector]:
        return self._selref_ptr_to_selector_map

    def selref_for_selector_name(self, selector_name: str) -> Optional[VirtualMemoryPointer]:
        return next(
            (selref for selref, selector in self._selref_ptr_to_selector_map.items() if selector.name == selector_name),
            None,
        )

    def get_method_imp_addresses(self, selector: str) -> List[VirtualMemoryPointer]:
        """Given a selector, return a list of virtual addresses corresponding to the start of each IMP for that SEL
        """
        return [
            objc_sel.implementation
            for objc_class in self.classes
            for objc_sel in objc_class.selectors
            if objc_sel.name == selector and objc_sel.implementation
        ]

    def objc_class_for_classlist_pointer(self, classlist_ptr: VirtualMemoryPointer) -> Optional[ObjcClass]:
        return self._classrefs_to_objc_classes.get(classlist_ptr)

    def _parse_objc_classes(self) -> List[ObjcClass]:
        """Read Objective-C class data in __objc_classlist, __objc_data to get classes and selectors in binary
        """
        logging.debug("Cross-referencing __objc_classlist, __objc_class, and __objc_data entries...")
        parsed_objc_classes = []
        classlist_pointers = self._get_classlist_pointers()
        for ptr in classlist_pointers:
            objc_class = self._get_objc_class_from_classlist_pointer(ptr)
            if objc_class:
                parsed_class = None
                # parse the instance method list
                objc_data_struct = self._get_objc_data_from_objc_class(objc_class)
                if objc_data_struct:
                    # the class's associated struct __objc_data contains the method list
                    parsed_class = self._parse_objc_data_entry(objc_class, objc_data_struct)

                # parse the metaclass if it exists
                # the class stores instance methods and the metaclass's method list contains class methods
                # the metaclass has the same name as the actual class
                metaclass = self._get_objc_class_from_classlist_pointer(VirtualMemoryPointer(objc_class.metaclass))
                if metaclass:
                    objc_data_struct = self._get_objc_data_from_objc_class(metaclass)
                    if objc_data_struct:
                        parsed_metaclass = self._parse_objc_data_entry(objc_class, objc_data_struct)
                        if parsed_class:
                            # add in selectors from the metaclass to the real class
                            parsed_class.selectors += parsed_metaclass.selectors
                        else:
                            # no base class found, set the base class to the metaclass
                            parsed_class = parsed_metaclass

                # sanity check
                # ensure we either found a class or metaclass
                if not parsed_class:
                    raise RuntimeError(f"Failed to parse classref {hex(ptr)}")
                parsed_objc_classes.append(parsed_class)
                self._classrefs_to_objc_classes[ptr] = parsed_class

        return parsed_objc_classes

    def _parse_objc_categories(self) -> List[ObjcCategory]:
        logging.debug("Cross referencing __objc_catlist, __objc_category, and __objc_data entries...")
        parsed_categories = []
        category_pointers = self._get_catlist_pointers()
        for ptr in category_pointers:
            objc_category_struct = self._get_objc_category_from_catlist_pointer(ptr)
            if objc_category_struct:
                parsed_category = self._parse_objc_category_entry(objc_category_struct)
                parsed_categories.append(parsed_category)
        return parsed_categories

    def _parse_class_and_category_info(self) -> List[ObjcClass]:
        """Parse classes and categories referenced by __objc_classlist and __objc_catlist
        """
        classes: List[ObjcClass] = []
        classes += self._parse_objc_classes()
        classes += self._parse_objc_categories()
        # Link superclass methods into their subclasses
        self._add_superclass_methods_to_subclasses()
        # Link superclasses of classes and base-classes of categories
        self._add_superclass_or_base_class_name_to_classes(classes)
        return classes

    def _add_superclass_methods_to_subclasses(self) -> None:
        for classlist_ptr, objc_class in self._classrefs_to_objc_classes.items():
            super_classref = objc_class.super_classref
            if not super_classref:
                continue
            superclass = self.objc_class_for_classlist_pointer(super_classref)
            if not superclass:
                continue
            objc_class.ivars += superclass.ivars
            objc_class.selectors += superclass.selectors
            objc_class.protocols += superclass.protocols

    def _add_superclass_or_base_class_name_to_classes(self, classes: List[ObjcClass]) -> None:
        """Iterate each ObjC class/category, and backfill its superclass/base_class name, respectively.

        Linking super/base_classes needs two data-sources, depending on whether the super/base_class is imported or not:
        - To retrieve the class names of imported super/base classes, this needs the map of bound dyld symbols
        - To retrieve the class names of locally implemented classes, this needs the full list of ObjcClasses
        """
        # For efficiency, build a map of (struct address -> class name)
        addr_to_class_names = {x.raw_struct.binary_offset: x.name for x in classes}

        for objc_class_or_category in classes:
            raw_struct = objc_class_or_category.raw_struct
            # This method uses the fact that `struct __objc_data.superclass` and `struct __objc_category.base_class`
            # have the same memory layout, being placed one 64-bit word after the start of the structure.
            base_class_field_addr = VirtualMemoryPointer(raw_struct.binary_offset + sizeof(c_uint64))

            # If the base class is an imported classref, the imported classref will be bound to its runtime load address
            # by dyld. Look up whether we have an import-binding for the `base_class` field of this structure.
            if base_class_field_addr in self.binary.dyld_bound_symbols:
                imported_base_class_sym = self.binary.dyld_bound_symbols[base_class_field_addr]
                base_class_name = imported_base_class_sym.name

            else:
                dereferenced_classref = VirtualMemoryPointer(self.binary.read_word(base_class_field_addr))
                # The base class is implemented in this binary, and we should have a corresponding ObjcClass object.
                if dereferenced_classref in addr_to_class_names:
                    base_class_name = addr_to_class_names[dereferenced_classref]
                else:
                    logging.error(
                        f"Failed to find a corresponding ObjC class for ref {dereferenced_classref} from "
                        f"{objc_class_or_category}"
                    )
                    base_class_name = "$_Unknown_Class"

            if isinstance(objc_class_or_category, ObjcCategory):
                objc_class_or_category.base_class = base_class_name
                # Update the name attribute to hold the parsed category name
                objc_class_or_category.name = f"{base_class_name} ({objc_class_or_category.category_name})"
            else:
                objc_class_or_category.superclass_name = base_class_name

    def _parse_global_protocol_info(self) -> List[ObjcProtocol]:
        """Parse protocols which code in the app conforms to, referenced by __objc_protolist
        """
        logging.debug("Cross referencing __objc_protolist, __objc_protocol, and __objc_data entries...")
        protocol_pointers = self._get_protolist_pointers()
        return self._parse_protocol_ptr_list(protocol_pointers)

    def read_ivars_from_ivarlist_ptr(self, ivarlist_ptr: VirtualMemoryPointer) -> List[ObjcIvar]:
        """Given the virtual address of an ivar list, return a List of each encoded ObjcIvar
        """
        ivarlist = self.binary.read_struct(ivarlist_ptr, ObjcIvarListStruct, virtual=True)
        ivars: List[ObjcIvar] = []
        # Parse each ivar struct which follows the ivarlist
        ivar_struct_ptr = ivarlist_ptr + ivarlist.sizeof
        for _ in range(ivarlist.count):
            ivar_struct = self.binary.read_struct_with_rebased_pointers(ivar_struct_ptr, ObjcIvarStruct, virtual=True)

            ivar_name = self.binary.get_full_string_from_start_address(ivar_struct.name)
            class_name = self.binary.get_full_string_from_start_address(ivar_struct.type)
            field_offset = self.binary.read_word(ivar_struct.offset_ptr, word_type=c_uint32)
            field_offset_addr = ivar_struct.offset_ptr

            # class_name and field_offset can be falsey ('' and 0), so don't include them in this sanity check
            if not ivar_name:
                raise ValueError(f"Failed to read ivar data for ivar entry @ {hex(ivar_struct_ptr)}")

            ivar = ObjcIvar(ivar_name, class_name, field_offset, field_offset_addr)  # type: ignore
            ivars.append(ivar)

            ivar_struct_ptr += ivar_struct.sizeof
        return ivars

    def read_selectors_from_methlist_ptr(self, methlist_ptr: VirtualMemoryPointer) -> List[ObjcSelector]:
        """Given the virtual address of a method list, return a List of ObjcSelectors encapsulating each method
        """
        methlist = self.binary.read_struct(methlist_ptr, ObjcMethodListStruct, virtual=True)
        selectors: List[ObjcSelector] = []
        # parse every entry in method list
        # the first entry appears directly after the ObjcMethodListStruct
        method_entry_off = methlist_ptr + methlist.sizeof
        for i in range(methlist.methcount):
            method_ent = ObjcMethodStruct.read_method_struct(
                self.binary, method_entry_off, methlist_flags=methlist.flags
            )
            # Byte-align IMP, as the lower bits are used for flags
            method_ent.implementation &= ~0x3  # type: ignore

            symbol_name = self.binary.get_full_string_from_start_address(method_ent.name)
            if not symbol_name:
                raise ValueError(f"Could not get symbol name for {method_ent.name}")
            # attempt to find corresponding selref
            selref = self._selector_literal_ptr_to_selref_map.get(method_ent.name)

            selector = ObjcSelector(symbol_name, selref, VirtualMemoryPointer(method_ent.implementation))
            selectors.append(selector)

            # save this selector in the selref pointer -> selector map
            if selref:
                # if this selector is already in the map, check if we now know the implementation address
                # we could have parsed the selector literal/selref pair in _parse_selrefs() but not have known the
                # implementation, but do now. It's also possible the selref is an external method, and thus will not
                # have a local implementation.
                most_specific_selector = selector
                if selref.source_address in self._selref_ptr_to_selector_map:
                    previously_parsed_selector = self._selref_ptr_to_selector_map[selref.source_address]
                    # Did we already parse this same selector but with more specific information?
                    # (Say, if we parse an ObjC class implementing a protocol before parsing the protocol itself)
                    if previously_parsed_selector.implementation:
                        # Make sure we keep the most specific selector we've seen
                        most_specific_selector = previously_parsed_selector
                self._selref_ptr_to_selector_map[selref.source_address] = most_specific_selector

            method_entry_off += method_ent.sizeof
        return selectors

    def _parse_objc_protocol_entry(self, objc_protocol_struct: ObjcProtocolRawStruct) -> ObjcProtocol:
        symbol_name = self.binary.get_full_string_from_start_address(objc_protocol_struct.name)
        if not symbol_name:
            raise ValueError(f"Could not get symbol name for {objc_protocol_struct.name}")

        selectors: List[ObjcSelector] = []
        if objc_protocol_struct.required_instance_methods:
            selectors += self.read_selectors_from_methlist_ptr(objc_protocol_struct.required_instance_methods)
        if objc_protocol_struct.required_class_methods:
            selectors += self.read_selectors_from_methlist_ptr(objc_protocol_struct.required_class_methods)
        if objc_protocol_struct.optional_instance_methods:
            selectors += self.read_selectors_from_methlist_ptr(objc_protocol_struct.optional_instance_methods)
        if objc_protocol_struct.optional_class_methods:
            selectors += self.read_selectors_from_methlist_ptr(objc_protocol_struct.optional_class_methods)

        return ObjcProtocol(objc_protocol_struct, symbol_name, selectors)

    def _parse_objc_category_entry(self, objc_category_struct: ObjcCategoryRawStruct) -> ObjcCategory:
        # TODO(PT): Add in the methods of the base class to the ObjcCategory
        symbol_name = self.binary.get_full_string_from_start_address(objc_category_struct.name)
        if not symbol_name:
            raise ValueError(f"Could not get symbol name for {objc_category_struct.name}")

        selectors: List[ObjcSelector] = []
        protocols: List[ObjcProtocol] = []

        # The class-name will be overwritten later in the parse. See self._add_superclass_or_base_class_name_to_classes
        placeholder_class_name = (
            f"<Base class of {symbol_name} category @ {objc_category_struct.binary_offset} will be populated later>"
        )

        # if the class implements no methods, the pointer to method list will be the null pointer
        # TODO(PT): we could add some flag to keep track of whether a given sel is an instance or class method
        if objc_category_struct.instance_methods:
            selectors += self.read_selectors_from_methlist_ptr(objc_category_struct.instance_methods)
        if objc_category_struct.class_methods:
            selectors += self.read_selectors_from_methlist_ptr(objc_category_struct.class_methods)
        if objc_category_struct.base_protocols:
            # TODO(PT): perhaps these should be combined into one call
            protocol_pointers = self._protolist_ptr_to_protocol_ptr_list(objc_category_struct.base_protocols)
            protocols += self._parse_protocol_ptr_list(protocol_pointers)

        return ObjcCategory(objc_category_struct, placeholder_class_name, symbol_name, selectors, protocols=protocols)

    def _parse_objc_data_entry(
        self, objc_class_struct: ObjcClassRawStruct, objc_data_struct: ObjcDataRawStruct
    ) -> ObjcClass:
        symbol_name = self.binary.get_full_string_from_start_address(objc_data_struct.name)
        if not symbol_name:
            raise ValueError(f"Could not get symbol name for {hex(objc_data_struct.name)}")

        selectors: List[ObjcSelector] = []
        protocols: List[ObjcProtocol] = []
        ivars: List[ObjcIvar] = []
        # if the class implements no methods, base_methods will be the null pointer
        if objc_data_struct.base_methods:
            selectors += self.read_selectors_from_methlist_ptr(objc_data_struct.base_methods)
        # if the class doesn't conform to any protocols, base_protocols will be null
        if objc_data_struct.base_protocols:
            protocol_pointer_list = self._protolist_ptr_to_protocol_ptr_list(objc_data_struct.base_protocols)
            protocols += self._parse_protocol_ptr_list(protocol_pointer_list)
        # Parse ivar list
        if objc_data_struct.ivars:
            ivars += self.read_ivars_from_ivarlist_ptr(objc_data_struct.ivars)

        return ObjcClass(objc_class_struct, symbol_name, selectors, ivars, protocols, objc_class_struct.superclass)

    def _protolist_ptr_to_protocol_ptr_list(self, protolist_ptr: VirtualMemoryPointer) -> List[VirtualMemoryPointer]:
        """Accepts the virtual address of an ObjcProtocolListStruct, and returns List of protocol pointers it refers to.
        """
        protolist = self.binary.read_struct(protolist_ptr, ObjcProtocolListStruct, virtual=True)
        protocol_pointers: List[VirtualMemoryPointer] = []
        # pointers start directly after the 'count' field
        addr = VirtualMemoryPointer(protolist.binary_offset + protolist.sizeof)
        for i in range(protolist.count):
            # This pointer may be rebased
            pointer = self.binary.read_rebased_pointer(addr)
            protocol_pointers.append(VirtualMemoryPointer(pointer))
            # step to next protocol pointer in list
            addr += sizeof(self.binary.platform_word_type)
        return protocol_pointers

    def _parse_protocol_ptr_list(self, protocol_ptrs: List[VirtualMemoryPointer]) -> List[ObjcProtocol]:
        protocols = []
        for protocol_ptr in protocol_ptrs:
            objc_protocol_struct = self._get_objc_protocol_from_pointer(protocol_ptr)
            if objc_protocol_struct:
                parsed_protocol = self._parse_objc_protocol_entry(objc_protocol_struct)
                protocols.append(parsed_protocol)
        return protocols

    def _get_catlist_pointers(self) -> List[VirtualMemoryPointer]:
        """Read pointers in __objc_catlist into list
        """
        _, catlist_pointers = self.binary.read_pointer_section("__objc_catlist")
        return catlist_pointers

    def _get_protolist_pointers(self) -> List[VirtualMemoryPointer]:
        """Read pointers in __objc_protolist into list
        """
        _, protolist_pointers = self.binary.read_pointer_section("__objc_protolist")
        return protolist_pointers

    def _get_classlist_pointers(self) -> List[VirtualMemoryPointer]:
        """Read pointers in __objc_classlist into list
        """
        _, classlist_pointers = self.binary.read_pointer_section("__objc_classlist")
        return classlist_pointers

    def _get_objc_category_from_catlist_pointer(
        self, category_struct_pointer: VirtualMemoryPointer
    ) -> ObjcCategoryRawStruct:
        """Read a struct __objc_category from the location indicated by the provided __objc_catlist pointer
        """
        category_entry = self.binary.read_struct_with_rebased_pointers(
            category_struct_pointer, ObjcCategoryRawStruct, virtual=True
        )
        return category_entry

    def _get_objc_protocol_from_pointer(self, protocol_struct_pointer: VirtualMemoryPointer) -> ObjcProtocolRawStruct:
        """Read a struct __objc_protocol from the location indicated by the provided struct objc_protocol_list pointer
        """
        protocol_entry = self.binary.read_struct_with_rebased_pointers(
            protocol_struct_pointer, ObjcProtocolRawStruct, virtual=True
        )
        return protocol_entry

    def _get_objc_class_from_classlist_pointer(self, class_struct_pointer: VirtualMemoryPointer) -> ObjcClassRawStruct:
        """Read a struct __objc_class from the location indicated by the __objc_classlist pointer
        """
        class_entry = self.binary.read_struct_with_rebased_pointers(
            class_struct_pointer, ObjcClassRawStruct, virtual=True
        )

        # sanitize class_entry
        # the least significant 2 bits are used for flags
        # flag 0x1 indicates a Swift class
        # mod data pointer to ignore flags!
        class_entry.data &= ~0x3  # type: ignore
        return class_entry

    def _get_objc_data_from_objc_class(self, objc_class: ObjcClassRawStruct) -> Optional[ObjcDataRawStruct]:
        """Read a struct __objc_data from a provided struct __objc_class
        If the struct __objc_class describe invalid or no corresponding data, None will be returned.
        """
        data_entry = self.binary.read_struct_with_rebased_pointers(objc_class.data, ObjcDataRawStruct, virtual=True)
        # ensure this is a valid entry
        if data_entry.name < self.binary.get_virtual_base():
            # TODO(PT): sometimes we'll get addresses passed to this method that are actually struct __objc_method
            # entries, rather than struct __objc_data entries. Investigate why this is.
            # This was observed on a 32bit binary, Esquire2
            logging.debug(
                f"caught ObjcDataRaw struct with invalid fields at {hex(int(objc_class.data))}."
                f" data->name = {hex(data_entry.name)}"
            )
            return None
        return data_entry



from enum import IntEnum


class MachoLoadCommands(IntEnum):
    # https://opensource.apple.com/source/cctools/cctools-949.0.1/include/mach-o/loader.h.auto.html
    LC_REQ_DYLD = 0x80000000

    LC_SEGMENT = 0x1
    LC_SYMTAB = 0x2
    LC_SYMSEG = 0x3
    LC_THREAD = 0x4
    LC_UNIXTHREAD = 0x5
    LC_LOADFVMLIB = 0x6
    LC_IDFVMLIB = 0x7
    LC_IDENT = 0x8
    LC_FVMFILE = 0x9
    LC_PREPAGE = 0xA
    LC_DYSYMTAB = 0xB
    LC_LOAD_DYLIB = 0xC
    LC_ID_DYLIB = 0xD
    LC_LOAD_DYLINKER = 0xE
    LC_ID_DYLINKER = 0xF
    LC_PREBOUND_DYLIB = 0x10

    LC_ROUTINES = 0x11
    LC_SUB_FRAMEWORK = 0x12
    LC_SUB_UMBRELLA = 0x13
    LC_SUB_CLIENT = 0x14
    LC_SUB_LIBRARY = 0x15
    LC_TWOLEVEL_HINTS = 0x16
    LC_PREBIND_CKSUM = 0x17

    LC_LOAD_WEAK_DYLIB = 0x18 | LC_REQ_DYLD

    LC_SEGMENT_64 = 0x19
    LC_ROUTINES_64 = 0x1A
    LC_UUID = 0x1B
    LC_RPATH = 0x1C | LC_REQ_DYLD
    LC_CODE_SIGNATURE = 0x1D
    LC_SEGMENT_SPLIT_INFO = 0x1E
    LC_REEXPORT_DYLIB = 0x1F | LC_REQ_DYLD
    LC_LAZY_LOAD_DYLIB = 0x20
    LC_ENCRYPTION_INFO = 0x21
    LC_DYLD_INFO = 0x22
    LC_DYLD_INFO_ONLY = 0x22 | LC_REQ_DYLD
    LC_LOAD_UPWARD_DYLIB = 0x23 | LC_REQ_DYLD
    LC_VERSION_MIN_MACOSX = 0x24
    LC_VERSION_MIN_IPHONEOS = 0x25
    LC_FUNCTION_STARTS = 0x26
    LC_DYLD_ENVIRONMENT = 0x27

    LC_MAIN = 0x28 | LC_REQ_DYLD
    LC_DATA_IN_CODE = 0x29
    LC_SOURCE_VERSION = 0x2A
    LC_DYLIB_CODE_SIGN_DRS = 0x2B
    LC_ENCRYPTION_INFO_64 = 0x2C
    LC_LINKER_OPTION = 0x2D
    LC_LINKER_OPTIMIZATION_HINT = 0x2E
    LC_VERSION_MIN_TVOS = 0x2F
    LC_VERSION_MIN_WATCHOS = 0x30
    LC_NOTE = 0x31
    LC_BUILD_VERSION = 0x32
    LC_DYLD_EXPORTS_TRIE = 0x33 | LC_REQ_DYLD
    LC_DYLD_CHAINED_FIXUPS = 0x34 | LC_REQ_DYLD



import re
from typing import List

from capstone import CsInsn
from capstone.arm64 import ARM64_OP_IMM, ARM64_OP_MEM, ARM64_OP_REG, Arm64Op

from strongarm.macho import (
    CPU_TYPE,
    MachoAnalyzer,
    MachoBinary,
    MachoParser,
    ObjcCategory,
    ObjcClass,
    ObjcSelector,
    VirtualMemoryPointer,
)
from strongarm.objc import (
    ObjcBranchInstruction,
    ObjcFunctionAnalyzer,
    ObjcInstruction,
    ObjcMethodInfo,
    RegisterContentsType,
)


class StringFormatter:
    @staticmethod
    def green(string: str) -> str:
        return f"\033[0;32m{string}\033[0m"

    @staticmethod
    def magenta(string: str) -> str:
        return StringFormatter.seed(197, string)

    @staticmethod
    def red(string: str) -> str:
        return f"\033[31;1m{string}\033[0m"

    @staticmethod
    def orange(string: str) -> str:
        return StringFormatter.seed(208, string)

    @staticmethod
    def blue(string: str) -> str:
        return f"\033[34;1m{string}\033[0m"

    @staticmethod
    def seed(seed: int, string: str) -> str:
        return f"\033[38;5;{seed}m{string}\033[0m"

    @staticmethod
    def none(string: str) -> str:
        return string

    @staticmethod
    def bold(string: str) -> str:
        return f"\033[1m{string}\033[0m"


def pick_macho_slice(parser: MachoParser) -> MachoBinary:
    """Retrieve a MachoBinary slice from a MachoParser, with a preference for an arm64 slice
    """
    binary_slices = parser.slices

    # Sanity checks (an empty list is falsey)
    if not binary_slices:
        raise ValueError(f"Could not parse {parser.path.name} as a Mach-O or FAT")

    parsed_binary = binary_slices[0]
    # Return 64 bit slice if there is one
    for binary_slice in binary_slices:
        parsed_binary = binary_slice
        if parsed_binary.cpu_type == CPU_TYPE.ARM64:
            break
    return parsed_binary


class _StringPalette:
    REG = StringFormatter.none
    IMM = StringFormatter.none
    MNEMONIC = StringFormatter.none
    BASIC_BLOCk = StringFormatter.none
    ADDRESS = StringFormatter.none
    ANNOTATION = StringFormatter.none
    STRING = StringFormatter.none


class StringPalette(_StringPalette):
    REG = StringFormatter.green
    IMM = StringFormatter.blue
    MNEMONIC = StringFormatter.magenta
    BASIC_BLOCK = StringFormatter.orange
    ADDRESS = StringFormatter.bold
    ANNOTATION = StringFormatter.orange
    ANNOTATION_ARGS = StringFormatter.blue
    STRING = StringFormatter.red


def format_instruction_arg(instruction: CsInsn, arg: Arm64Op) -> str:
    if arg.type == ARM64_OP_REG:
        return StringPalette.REG(instruction.reg_name(arg.value.reg))
    elif arg.type == ARM64_OP_IMM:
        return StringPalette.IMM(hex(arg.value.imm))
    elif arg.type == ARM64_OP_MEM:
        return f"[{StringPalette.REG(instruction.reg_name(arg.mem.base))} #{StringPalette.IMM(hex(arg.mem.disp))}]"
    raise RuntimeError(f"unknown arg type {arg.type}")


def args_from_sel_name(sel: str) -> List[str]:
    sel_args = ["self", f"@selector({sel})"]
    if ":" not in sel:
        return sel_args

    sel_components = sel.split(":")
    for component in sel_components:
        if not len(component):
            sel_args.append("")
            continue
        # extract the last capitalized word
        split = re.findall("[A-Z][^A-Z]*", component)
        # if no capitalized word, use the full component
        if not len(split):
            split.append(component)
        # lowercase it
        sel_args.append(split[-1].lower())
    return sel_args


def disassemble_method(binary: MachoBinary, method: ObjcMethodInfo) -> str:
    disassembled_text: List[str] = []

    # Figure out the arguments based on the sel name
    sel_args = args_from_sel_name(method.objc_sel.name)

    argument_list = ", ".join(sel_args)
    signature = f"\n-[{method.objc_class.name} {method.objc_sel.name}]({argument_list});"
    disassembled_text.append(signature)

    if not method.imp_addr:
        return f"Could not find address for [{method.objc_class.name} {method.objc_sel.name}]"
    return disassemble_function(binary, method.imp_addr, disassembled_text, sel_args)


def print_instr(instr: ObjcInstruction) -> None:
    raw_instr = instr.raw_instr
    instruction_string = f"\t{hex(instr.address)}\t\t{raw_instr.mnemonic}"

    # Add each arg to the string
    instruction_string += ", ".join([format_instruction_arg(raw_instr, arg) for arg in raw_instr.operands])
    print(instruction_string)


def annotate_instruction(function_analyzer: ObjcFunctionAnalyzer, sel_args: List[str], instr: CsInsn) -> str:
    annotation = "\t\t"
    # Parse as an ObjcInstruction
    wrapped_instr = ObjcInstruction.parse_instruction(
        function_analyzer, function_analyzer.get_instruction_at_address(instr.address)
    )

    if isinstance(wrapped_instr, ObjcBranchInstruction):
        wrapped_branch_instr: ObjcBranchInstruction = wrapped_instr

        annotation += "#\t"
        if function_analyzer.is_local_branch(wrapped_branch_instr):
            annotation += StringPalette.ANNOTATION(f"jump loc_{hex(wrapped_branch_instr.destination_address)}")

        elif wrapped_instr.symbol:
            annotation += StringPalette.ANNOTATION(wrapped_instr.symbol)

            if not wrapped_branch_instr.selector:
                annotation += StringPalette.ANNOTATION("();")
            else:
                annotation += StringPalette.ANNOTATION_ARGS(f"(id, @selector({wrapped_branch_instr.selector.name})")

                # Figure out argument count passed to selector
                arg_count = wrapped_branch_instr.selector.name.count(":")
                for i in range(arg_count):
                    # x0 is self, x1 is the SEL, real args start at x2
                    register = f"x{i + 2}"
                    method_arg = function_analyzer.get_register_contents_at_instruction(register, wrapped_branch_instr)

                    method_arg_string = ", "
                    if method_arg.type == RegisterContentsType.IMMEDIATE:
                        method_arg_string += hex(method_arg.value)
                    else:
                        method_arg_string += "<?>"

                    annotation += StringPalette.STRING(method_arg_string)
                annotation += ");"

        else:
            annotation += StringPalette.ANNOTATION(f"({hex(instr.address)})(")
            arg_count = 4
            for i in range(arg_count):
                # x0 is self, x1 is the SEL, real args start at x2
                register = f"x{i}"
                method_arg = function_analyzer.get_register_contents_at_instruction(register, wrapped_instr)

                method_arg_string = f"{register}: "
                if method_arg.type == RegisterContentsType.IMMEDIATE:
                    method_arg_string += hex(method_arg.value)
                else:
                    method_arg_string += "<?>"

                annotation += StringPalette.ANNOTATION_ARGS(method_arg_string)
                annotation += ", "
            annotation += ");"
    else:
        # Try to annotate string loads
        # This code taken from Ethan's potential passwords check
        if instr.mnemonic in ["ldr", "adr", "adrp", "add"]:
            # Only care about general purpose registers that are being written into
            if not ObjcInstruction.instruction_uses_vector_registers(instr):
                _, instr_mutated_regs = instr.regs_access()
                if len(instr_mutated_regs):
                    # Get the contents of the register (an address)
                    register = instr.reg_name(instr_mutated_regs[0])
                    wrapped_instr = ObjcInstruction.parse_instruction(function_analyzer, instr)
                    register_contents = function_analyzer.get_register_contents_at_instruction(register, wrapped_instr)
                    if register_contents.type == RegisterContentsType.IMMEDIATE:
                        # Try reading a string
                        binary_str = function_analyzer.binary.read_string_at_address(
                            VirtualMemoryPointer(register_contents.value)
                        )
                        if binary_str:
                            annotation += StringPalette.STRING(f'#\t"{binary_str}"')

    return annotation


def disassemble_function(
    binary: MachoBinary, function_addr: VirtualMemoryPointer, prefix: List[str] = None, sel_args: List[str] = None
) -> str:
    if not prefix:
        prefix = []
    if not sel_args:
        sel_args = []

    disassembled_text = prefix
    function_analyzer = ObjcFunctionAnalyzer.get_function_analyzer(binary, function_addr)

    # Transform basic blocks into tuples of (basic block start addr, basic block end addr)
    basic_block_boundaries = [[block.start_address, block.end_address] for block in function_analyzer.basic_blocks]
    # Flatten basic_block_boundaries into one-dimensional list
    basic_block_boundaries_flat = [x for boundaries in basic_block_boundaries for x in boundaries]
    # Remove duplicate boundaries
    basic_block_boundaries_set = set(basic_block_boundaries_flat)

    for instr in function_analyzer.instructions:
        line = ""
        # Add visual indicator if this is a basic block boundary
        if instr.address in basic_block_boundaries_set:
            line += StringPalette.BASIC_BLOCK(f"--- loc_{hex(instr.address)} ----------\n")

        lines = [
            StringPalette.ADDRESS(hex(instr.address)),
            StringPalette.MNEMONIC(f"{instr.mnemonic:5}"),
            ", ".join([format_instruction_arg(instr, x) for x in instr.operands]),
            annotate_instruction(function_analyzer, sel_args, instr),
        ]
        line += "\t" + "\t".join(lines)
        disassembled_text.append(line)

    return "\n".join(disassembled_text)


def print_binary_info(binary: MachoBinary) -> None:
    print(f"Mach-O type: {binary.file_type.name}")
    print(f"{'Big' if binary.is_swap else 'Little'} endian")
    print(f"Virtual base: {hex(binary.get_virtual_base())}")
    print(f"Contains encrypted section? {binary.is_encrypted()}")


def print_binary_load_commands(binary: MachoBinary) -> None:
    print("\nLoad commands:")
    load_commands = binary.load_dylib_commands
    for cmd in load_commands:
        dylib_name_addr = binary.get_virtual_base() + cmd.binary_offset + cmd.dylib.name.offset
        dylib_name = binary.read_string_at_address(dylib_name_addr)
        dylib_version = cmd.dylib.current_version
        print(f"\t{dylib_name} v.{hex(dylib_version)}")


def print_binary_segments(binary: MachoBinary) -> None:
    print("\nSegments:")
    for segment in binary.segments:
        virtual_loc = f"[{segment.vmaddr:#011x} - {segment.vmaddr + segment.vmsize:#011x}]"
        file_loc = f"[{segment.offset:#011x} - {segment.offset + segment.size:#011x}]"
        print(f"\t{virtual_loc} (file {file_loc}) {segment.name}")


def print_binary_sections(binary: MachoBinary) -> None:
    print("\nSections:")
    for section in binary.sections:
        print(f"\t[{hex(section.address)} - {hex(section.end_address)}] {section.name} ({section.segment.name})")


def print_analyzer_imported_symbols(analyzer: MachoAnalyzer) -> None:
    print("\nSymbols:")
    print("\tImported symbols:")
    stub_map = analyzer.imported_symbol_names_to_pointers
    for imported_sym in analyzer.imported_symbols:
        print(f"\t\t{imported_sym}: ", end="")
        # Attempt to find the call stub for this symbol
        stub_location = ""
        if imported_sym in stub_map:
            stub_location = f"dyld stub at {hex(stub_map[imported_sym])}"
        print(stub_location)


def print_analyzer_exported_symbols(analyzer: MachoAnalyzer) -> None:
    print("\tExported symbols:")
    for exported_sym, exported_addr in analyzer.exported_symbol_names_to_pointers.items():
        print(f"\t\t{exported_sym}: {hex(exported_addr)}")


def print_selector(objc_class: ObjcClass, selector: ObjcSelector) -> None:
    # Belongs to a class or category?
    if isinstance(objc_class, ObjcCategory):
        category: ObjcCategory = objc_class
        class_name = f"{category.base_class} ({category.name})"
    else:
        class_name = objc_class.name
    if selector.implementation:
        print(f"\t-[{class_name} {selector.name}] defined at {hex(selector.implementation)}")
    else:
        print(f"\t-[{class_name} {selector.name}]")


def print_analyzer_methods(analyzer: MachoAnalyzer) -> None:
    print("\nObjective-C Methods:")
    methods = analyzer.get_objc_methods()
    for method_info in methods:
        print_selector(method_info.objc_class, method_info.objc_sel)


def print_analyzer_classes(analyzer: MachoAnalyzer) -> None:
    print("\nObjective-C Classes:")
    classes = analyzer.objc_classes()
    classes = sorted(classes, key=lambda c: c.name)
    for objc_class in classes:
        # Belongs to a class or category?
        if isinstance(objc_class, ObjcCategory):
            category: ObjcCategory = objc_class
            class_name = f"{category.base_class} ({category.name})"
        else:
            class_name = objc_class.name
        print(f"\t{class_name}: {len(objc_class.selectors)} selectors")


def print_analyzer_protocols(analyzer: MachoAnalyzer) -> None:
    print("\nProtocols conformed to within the binary:")
    protocols = analyzer.get_conformed_protocols()
    protocols = sorted(protocols, key=lambda p: p.name)
    for protocol in protocols:
        print(f"\t{protocol.name}: {len(protocol.selectors)} selectors")


def print_raw_strings(binary: MachoBinary) -> None:
    strings_section = binary.section_with_name("__cstring", "__TEXT")
    if strings_section is None:
        return

    print("\nStrings:")
    strings_content = binary.get_bytes(strings_section.offset, strings_section.size)
    for string in strings_content.split(b"\0"):
        try:
            print(f"\t{string.decode()}")
        except UnicodeDecodeError:
            print(f"\t{string}")



import logging
from ctypes import Structure, c_uint64, sizeof
from distutils.version import LooseVersion
from typing import TYPE_CHECKING, Any, Optional, Type, Union

from strongarm.macho.macho_definitions import (
    CFString32,
    CFString64,
    DylibCommand,
    MachoBuildToolVersion,
    MachoBuildVersionCommand,
    MachoDyldChainedFixupsHeaderRaw,
    MachoDyldChainedImportRaw,
    MachoDyldChainedPtr64BindRaw,
    MachoDyldChainedPtr64RebaseRaw,
    MachoDyldChainedStartsInImageRaw,
    MachoDyldChainedStartsInSegmentRaw,
    MachoDyldInfoCommand,
    MachoDysymtabCommand,
    MachoEncryptionInfo32Command,
    MachoEncryptionInfo64Command,
    MachoHeader32,
    MachoHeader64,
    MachoLinkeditDataCommand,
    MachoLoadCommand,
    MachoNlist32,
    MachoNlist64,
    MachoSection32Raw,
    MachoSection64Raw,
    MachoSegmentCommand32,
    MachoSegmentCommand64,
    MachoSymtabCommand,
    ObjcCategoryRaw32,
    ObjcCategoryRaw64,
    ObjcClassRaw32,
    ObjcClassRaw64,
    ObjcDataRaw32,
    ObjcDataRaw64,
    ObjcIvar32,
    ObjcIvar64,
    ObjcIvarList,
    ObjcMethod32,
    ObjcMethod64,
    ObjcMethodList,
    ObjcMethodRelativeData,
    ObjcProtocolList32,
    ObjcProtocolList64,
    ObjcProtocolRaw32,
    ObjcProtocolRaw64,
    VirtualMemoryPointer,
)

# create type alias for the following classes that inherit from ArchIndependentStructure
if TYPE_CHECKING:
    from .codesign.codesign_definitions import (  # noqa: F401
        CSBlobIndexStruct,
        CSBlobStruct,
        CSCodeDirectoryStruct,
        CSSuperblobStruct,
    )
    from .macho_binary import MachoBinary

# Create type alias for the following classes that inherit from ArchIndependentStructure
_32_BIT_STRUCT_ALIAS = Union[
    Type[MachoHeader32],
    Type[MachoSegmentCommand32],
    Type[MachoSection32Raw],
    Type[MachoEncryptionInfo32Command],
    Type[MachoNlist32],
    Type[MachoLoadCommand],
    Type[MachoSymtabCommand],
    Type[MachoDysymtabCommand],
    Type[MachoDyldInfoCommand],
    Type[MachoLinkeditDataCommand],
    Type[ObjcDataRaw32],
    Type[ObjcClassRaw32],
    Type[ObjcMethod32],
    Type[ObjcIvar32],
    Type[ObjcMethodList],
    Type[ObjcCategoryRaw32],
    Type[ObjcProtocolRaw32],
    Type[ObjcProtocolList32],
    Type[CFString32],
    Type[DylibCommand],
    Type["CSBlobStruct"],
    Type["CSSuperblobStruct"],
    Type["CSCodeDirectoryStruct"],
    Type["CSBlobIndexStruct"],
    Type["ObjcIvarList"],
    Type["MachoBuildVersionCommand"],
    Type["MachoBuildToolVersion"],
]

_64_BIT_STRUCT_ALIAS = Union[
    Type[MachoHeader64],
    Type[MachoSegmentCommand64],
    Type[MachoSection64Raw],
    Type[MachoEncryptionInfo64Command],
    Type[MachoNlist64],
    Type[MachoLoadCommand],
    Type[MachoSymtabCommand],
    Type[MachoDysymtabCommand],
    Type[MachoDyldInfoCommand],
    Type[MachoLinkeditDataCommand],
    Type[ObjcDataRaw64],
    Type[ObjcClassRaw64],
    Type[ObjcMethod64],
    Type[ObjcIvar64],
    Type[ObjcMethodList],
    Type[ObjcCategoryRaw64],
    Type[ObjcProtocolRaw64],
    Type[ObjcProtocolList64],
    Type[CFString64],
    Type[DylibCommand],
    Type["CSBlobStruct"],
    Type["CSSuperblobStruct"],
    Type["CSCodeDirectoryStruct"],
    Type["CSBlobIndexStruct"],
    Type["ObjcIvarList"],
    Type["MachoBuildVersionCommand"],
    Type["MachoBuildToolVersion"],
    Type["MachoDyldChainedFixupsHeaderRaw"],
    Type["MachoDyldChainedImportRaw"],
    Type["MachoDyldChainedStartsInImageRaw"],
    Type["MachoDyldChainedStartsInSegmentRaw"],
    Type["MachoDyldChainedPtr64RebaseRaw"],
    Type["MachoDyldChainedPtr64BindRaw"],
    Type["MachoBuildToolVersion"],
]


class ArchIndependentStructure:
    _32_BIT_STRUCT: Optional[_32_BIT_STRUCT_ALIAS] = None
    _64_BIT_STRUCT: Optional[_64_BIT_STRUCT_ALIAS] = None

    @classmethod
    def get_backing_data_layout(
        cls, is_64bit: bool = True, minimum_deployment_target: Optional[LooseVersion] = None
    ) -> Type[Structure]:
        """The underlying data layout may be different depending on the binary type.
        Args:
            is_64bit: Binary's 64 bitness
            minimum_deployment_target: The minimum deployment target (target OS version) of the binary, if available
        Returns:
            size of the structure in bytes
        """
        struct_type = cls._64_BIT_STRUCT if is_64bit else cls._32_BIT_STRUCT

        if struct_type is None:
            raise ValueError("Undefined struct_type")

        return struct_type

    def __init__(self, binary_offset: int, struct_bytes: bytearray, backing_layout: Type[Structure]):
        struct: ArchIndependentStructure = backing_layout.from_buffer(struct_bytes)  # type: ignore

        for field_name, *_ in struct._fields_:
            # clone fields from struct to this class
            setattr(self, field_name, getattr(struct, field_name))

        # record size of underlying struct, for when traversing file by structs
        self.sizeof = sizeof(backing_layout)
        # record the location in the binary this struct was parsed from
        self.binary_offset = binary_offset

    if TYPE_CHECKING:
        # GVR suggested to use this pattern to ignore dynamic attribute assignment errors
        def __getattr__(self, key: str) -> Any:
            pass

    def __repr__(self) -> str:
        attributes = "\t".join([f"{x}: {getattr(self, x)}" for x in self.__dict__.keys()])
        rep = f"{self.__class__.__name__} ({attributes})"
        return rep


class MachoHeaderStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoHeader32
    _64_BIT_STRUCT = MachoHeader64


class MachoSegmentCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoSegmentCommand32
    _64_BIT_STRUCT = MachoSegmentCommand64


class MachoSectionRawStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoSection32Raw
    _64_BIT_STRUCT = MachoSection64Raw


class MachoEncryptionInfoStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoEncryptionInfo32Command
    _64_BIT_STRUCT = MachoEncryptionInfo64Command


class MachoNlistStruct(ArchIndependentStructure):
    __slots__ = ["n_un", "n_type", "n_sect", "n_desc", "n_value"]
    _32_BIT_STRUCT = MachoNlist32
    _64_BIT_STRUCT = MachoNlist64


class ObjcDataRawStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcDataRaw32
    _64_BIT_STRUCT = ObjcDataRaw64


class ObjcProtocolRawStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcProtocolRaw32
    _64_BIT_STRUCT = ObjcProtocolRaw64


class ObjcProtocolListStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcProtocolList32
    _64_BIT_STRUCT = ObjcProtocolList64


class ObjcCategoryRawStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcCategoryRaw32
    _64_BIT_STRUCT = ObjcCategoryRaw64


class ObjcClassRawStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcClassRaw32
    _64_BIT_STRUCT = ObjcClassRaw64


class ObjcMethodStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcMethod32
    _64_BIT_STRUCT = ObjcMethod64

    @classmethod
    def get_backing_data_layout(
        cls,
        is_64bit: bool = True,
        minimum_deployment_target: Optional[LooseVersion] = None,
        methlist_flags: Optional[int] = None,
    ) -> Type[Structure]:
        # Prior to iOS 14, 64-bit targets would use an ObjcMethod64 structure with absolute addresses.
        # On iOS 14 and later, 64-bit targets use a structure with 32-bit relative offsets from each field.
        if is_64bit and minimum_deployment_target and minimum_deployment_target >= LooseVersion("14.0.0"):
            # SCAN-2419: Binaries can be built for iOS 14 and still use an absolute method list, so also check a flag
            # bit set in the method list
            if methlist_flags and methlist_flags & (1 << 31) != 0:
                return ObjcMethodRelativeData

        return super().get_backing_data_layout(is_64bit, minimum_deployment_target)

    @classmethod
    def read_method_struct(
        cls, binary: "MachoBinary", address: VirtualMemoryPointer, methlist_flags: Optional[int] = None
    ) -> "ObjcMethodStruct":  # noqa
        """Read an ObjcMethodStruct from the provided binary address.
        This method accounts for post-iOS-14 binaries using a relative-offset layout for this structure, and
         patches the field values to appear as absolute addresses to callers, to match the layout from prior versions.
        """
        struct_type = cls.get_backing_data_layout(
            binary.is_64bit, binary.get_minimum_deployment_target(), methlist_flags
        )
        data = binary.get_contents_from_address(address=address, size=sizeof(struct_type), is_virtual=True)
        method_ent = ObjcMethodStruct(address, data, struct_type)

        # If we're parsing the iOS14+ structure that encodes signed 32b offsets instead of 64b absolute addresses,
        # translate the offsets to absolute addresses for caller convenience.
        if struct_type == ObjcMethodRelativeData:
            # Fix up each field by translating it from a 32b signed offset to an absolute address
            method_entry_off = address
            method_ent.signature += method_entry_off + 4  # type: ignore
            method_ent.implementation += method_entry_off + 8  # type: ignore

            # Rather than pointing to a selector literal, this field points to a selref. Dereference it now
            selref_addr = method_ent.name + method_entry_off  # type: ignore
            # This selref may be rebased
            method_ent.name = binary.read_rebased_pointer(selref_addr)  # type: ignore
        else:
            for field_name, field_type, *_ in struct_type._fields_:
                field_offset = getattr(getattr(struct_type, field_name), "offset")
                field_address = address + field_offset
                if field_type == c_uint64 and field_address in binary.dyld_rebased_pointers:
                    pointer_value = binary.dyld_rebased_pointers[field_address]
                    logging.debug(
                        f"Setting rebased pointer within {struct_type}+{field_offset} -> "
                        f"{pointer_value} at {field_address}"
                    )
                    setattr(method_ent, field_name, pointer_value)

        return method_ent


class ObjcIvarStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcIvar32
    _64_BIT_STRUCT = ObjcIvar64


class CFStringStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = CFString32
    _64_BIT_STRUCT = CFString64


class ObjcMethodListStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcMethodList
    _64_BIT_STRUCT = ObjcMethodList


class ObjcIvarListStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = ObjcIvarList
    _64_BIT_STRUCT = ObjcIvarList


class DylibCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = DylibCommand
    _64_BIT_STRUCT = DylibCommand


class MachoLoadCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoLoadCommand
    _64_BIT_STRUCT = MachoLoadCommand


class MachoSymtabCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoSymtabCommand
    _64_BIT_STRUCT = MachoSymtabCommand


class MachoDysymtabCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoDysymtabCommand
    _64_BIT_STRUCT = MachoDysymtabCommand


class MachoDyldInfoCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoDyldInfoCommand
    _64_BIT_STRUCT = MachoDyldInfoCommand


class MachoLinkeditDataCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoLinkeditDataCommand
    _64_BIT_STRUCT = MachoLinkeditDataCommand


class MachoBuildVersionCommandStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoBuildVersionCommand
    _64_BIT_STRUCT = MachoBuildVersionCommand


class MachoBuildToolVersionStruct(ArchIndependentStructure):
    _32_BIT_STRUCT = MachoBuildToolVersion
    _64_BIT_STRUCT = MachoBuildToolVersion


class MachoDyldChainedFixupsHeader(ArchIndependentStructure):
    _64_BIT_STRUCT = MachoDyldChainedFixupsHeaderRaw


class MachoDyldChainedImport(ArchIndependentStructure):
    _64_BIT_STRUCT = MachoDyldChainedImportRaw


class MachoDyldChainedStartsInImage(ArchIndependentStructure):
    _64_BIT_STRUCT = MachoDyldChainedStartsInImageRaw


class MachoDyldChainedStartsInSegment(ArchIndependentStructure):
    _64_BIT_STRUCT = MachoDyldChainedStartsInSegmentRaw


class MachoDyldChainedPtr64Rebase(ArchIndependentStructure):
    _64_BIT_STRUCT = MachoDyldChainedPtr64RebaseRaw


class MachoDyldChainedPtr64Bind(ArchIndependentStructure):
    _64_BIT_STRUCT = MachoDyldChainedPtr64BindRaw



from ctypes import c_uint32, sizeof
from pathlib import Path
from typing import List, Optional

from strongarm.macho.macho_binary import MachoBinary
from strongarm.macho.macho_definitions import MachArch, MachoFatArch, MachoFatHeader, StaticFilePointer, swap32


class ArchitectureNotSupportedError(Exception):
    pass


class MachoParser:
    _FAT_MAGIC = [MachArch.FAT_MAGIC, MachArch.FAT_CIGAM]
    _MACHO_MAGIC = [MachArch.MH_MAGIC, MachArch.MH_CIGAM, MachArch.MH_MAGIC_64, MachArch.MH_CIGAM_64]
    _BIG_ENDIAN_MAG = [MachArch.FAT_CIGAM, MachArch.MH_CIGAM, MachArch.MH_CIGAM_64]

    _SUPPORTED_SLICE_MAG = MachoBinary.SUPPORTED_MAG

    SUPPORTED_MAG = _FAT_MAGIC + _SUPPORTED_SLICE_MAG

    def __init__(self, path: Path) -> None:
        self.path = path

        self.header: Optional[MachoFatHeader] = None
        self.is_swapped: bool = False
        self.slices: List[MachoBinary] = []

        self.parse()

    def get_arm64_slice(self) -> Optional[MachoBinary]:
        """Retrieve the parsed slice from the FAT built for ARM64
        """
        arm64_slices = [x for x in self.slices if x.header.cputype == MachArch.MH_CPU_TYPE_ARM64]
        if len(arm64_slices):
            return arm64_slices[0]
        return None

    def get_armv7_slice(self) -> Optional[MachoBinary]:
        """Retrieve the parsed slice from the FAT built for ARMv7
        """
        armv7_slices = [x for x in self.slices if x.header.cputype == MachArch.MH_CPU_TYPE_ARM]
        if len(armv7_slices):
            return armv7_slices[0]
        return None

    def parse(self) -> None:
        """Parse a Mach-O or FAT archive represented by file at a given path
        This method will throw an exception if an binary is passed which is malformed or not a
        valid Mach-O or FAT archive
        """
        if not self.is_magic_supported():
            raise ArchitectureNotSupportedError(self.path.as_posix())

        self.is_swapped = self.should_swap_bytes()

        if self.is_fat:
            self.parse_fat_header()
        else:
            file_size = self.path.stat().st_size
            self.parse_thin_header(StaticFilePointer(0), file_size)

    def parse_thin_header(self, fileoff: StaticFilePointer, slice_size: int) -> None:
        """Parse a known Mach-O header at a given file offset, and add it to self.slices
        This method will throw an Exception if the data at fileoff is not a valid Mach-O header

        Args:
            fileoff: byte index into file to interpret Mach-O header at
            slice_size: Byte-count of the Mach-O slice in the file
        """
        # sanity check
        if not self._check_is_macho_header(fileoff):
            raise RuntimeError(f"Parsing error: data at file offset {hex(int(fileoff))} was not a valid Mach-O slice!")

        slice_data = self.get_bytes(fileoff, slice_size)
        attempt = MachoBinary(self.path, slice_data, file_offset=fileoff)

        # if the MachoBinary does not have a header, there was a problem parsing it
        if not attempt.header:
            raise RuntimeError("parsed MachoBinary missing Mach-O header field")
        self.slices.append(attempt)

    def parse_fat_header(self) -> None:
        """Parse the FAT header implicitly found at the start of the file
        This method will also parse all Mach-O slices that the FAT describes
        """
        # sanity check
        if self._check_is_macho_header(StaticFilePointer(0)):
            raise RuntimeError("Parsing error: Expected FAT header but found incorrect magic!")

        # start reading from the start of the file
        read_off = 0
        header_bytes = bytearray(self.get_bytes(StaticFilePointer(read_off), sizeof(MachoFatHeader)))
        self.header = MachoFatHeader.from_buffer(header_bytes)
        # first fat_arch structure is directly after FAT header
        read_off += sizeof(MachoFatHeader)

        # remember to swap fields if file contains non-native byte order
        if self.is_swapped:
            self.header.nfat_arch = swap32(self.header.nfat_arch)  # type: ignore

        for i in range(self.header.nfat_arch):  # type: ignore
            arch_bytes = self.get_bytes(StaticFilePointer(read_off), sizeof(MachoFatArch))
            fat_arch = MachoFatArch.from_buffer(bytearray(arch_bytes))

            # do we need to byte swap?
            # TODO(pt): come up with more elegant mechanism for swapping byte order in every word of Structure
            if self.is_swapped:
                # non-native byte order, swap every field in fat_arch
                fat_arch.cputype = swap32(int(fat_arch.cputype))
                fat_arch.cpusubtype = swap32(int(fat_arch.cpusubtype))
                fat_arch.offset = swap32(int(fat_arch.offset))
                fat_arch.size = swap32(int(fat_arch.size))
                fat_arch.align = swap32(int(fat_arch.align))

            # TODO(PT): SCAN-2910: Don't skip ARM64E slices
            if fat_arch.cputype == MachArch.MH_CPU_TYPE_ARM64 and fat_arch.cpusubtype == (
                MachArch.MH_CPU_SUBTYPE_PTRAUTH_ABI | MachArch.MH_CPU_SUBTYPE_ARM64E
            ):
                continue

            self.parse_thin_header(StaticFilePointer(fat_arch.offset), fat_arch.size)
            # move to next fat_arch structure in file
            read_off += sizeof(MachoFatArch)

    def _check_is_macho_header(self, offset: StaticFilePointer) -> bool:
        """Check if the data located at a file offset represents a valid Mach-O header, based on the magic

        Args:
            offset: File offset to read magic from

        Returns:
            True if the byte content of the file at 'offset' contain the magic number for a Mach-O slice,
            False if the magic is anything else

        """
        magic = c_uint32.from_buffer(bytearray(self.get_bytes(offset, sizeof(c_uint32)))).value
        return magic in MachoParser._MACHO_MAGIC

    def is_magic_supported(self) -> bool:
        """Check whether a magic number represents a file format which this class is capable of parsing

        Args:
            magic: Magic value denoting file type to check against

        Returns:
            True if the magic number represents a supported file format, False otherwise

        """
        return self.file_magic in MachoParser.SUPPORTED_MAG

    @property
    def file_magic(self) -> int:
        """Read file magic
        """
        return c_uint32.from_buffer(bytearray(self.get_bytes(StaticFilePointer(0), sizeof(c_uint32)))).value

    @property
    def is_fat(self) -> bool:
        """Check if file magic indicates a FAT archive or not

        Returns:
            True if the magic indicates FAT format, False otherwise

        """
        return self.file_magic in MachoParser._FAT_MAGIC

    def should_swap_bytes(self) -> bool:
        """Check if we need to swap due to a difference in endianness between host and binary

        Returns:
            True if the host and binary differ in endianness, False otherwise

        """
        # TODO(PT): figure out whether we need to swap to little or big endian,
        # based on system endianness and binary endianness
        # everything we touch currently is little endian, so let's not worry about it for now
        return self.file_magic in MachoParser._BIG_ENDIAN_MAG

    def get_bytes(self, offset: StaticFilePointer, size: int) -> bytes:
        """Read a byte list from binary file of a given size, starting from a given offset

        Args:
            offset: Offset within file to begin reading from
            size: Maximum number of bytes to read

        Returns:
            Byte list representing contents of file at provided address

        """
        with open(self.path, "rb") as binary_file:
            binary_file.seek(offset)
            return binary_file.read(size)



import logging
from ctypes import c_long, c_uint16, c_uint32, c_uint64, sizeof
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Dict, List, Optional, Tuple

from .arch_independent_structs import (
    DylibCommandStruct,
    MachoDyldChainedFixupsHeader,
    MachoDyldChainedImport,
    MachoDyldChainedPtr64Bind,
    MachoDyldChainedPtr64Rebase,
    MachoDyldChainedStartsInImage,
    MachoDyldChainedStartsInSegment,
)
from .macho_binary import MachoBinary
from .macho_definitions import StaticFilePointer, VirtualMemoryPointer


class BindOpcode(IntEnum):
    BIND_OPCODE_DONE = 0x00
    BIND_OPCODE_SET_DYLIB_ORDINAL_IMM = 0x10
    BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB = 0x20
    BIND_OPCODE_SET_DYLIB_SPECIAL_IMM = 0x30
    BIND_OPCODE_SET_SYMBOL_TRAILING_FLAGS_IMM = 0x40
    BIND_OPCODE_SET_TYPE_IMM = 0x50
    BIND_OPCODE_SET_ADDEND_SLEB = 0x60
    BIND_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB = 0x70
    BIND_OPCODE_ADD_ADDR_ULEB = 0x80
    BIND_OPCODE_DO_BIND = 0x90
    BIND_OPCODE_DO_BIND_ADD_ADDR_ULEB = 0xA0
    BIND_OPCODE_DO_BIND_ADD_ADDR_IMM_SCALED = 0xB0
    BIND_OPCODE_DO_BIND_ULEB_TIMES_SKIPPING_ULEB = 0xC0
    BIND_OPCODE_THREADED = 0xD0

    # The immediate will contain a sub-opcode for BIND_OPCODE_THREADED
    BIND_SUBOPCODE_THREADED_SET_BIND_ORDINAL_TABLE_SIZE_ULEB = 0x00
    BIND_SUBOPCODE_THREADED_APPLY = 0x01


@dataclass
class DyldBoundSymbol:
    binary: MachoBinary
    address: VirtualMemoryPointer
    library_ordinal: int
    name: str
    dylib: Optional[DylibCommandStruct] = field(init=False)

    def __post_init__(self) -> None:
        self.dylib = self.binary.dylib_for_library_ordinal(self.library_ordinal)


class DyldInfoParser:
    """Wraps up the logic to parse __LINKEDIT data so that we can make sense of rebased pointers and bound dyld symbols.
    On < iOS 15 binaries:
        Parses the dyld bytecode stream and extract dyld bound addresses to the DyldBoundSymbol they represent
    On >= iOS 15 binaries:
        Parses the chained fixup pointers and generates a map of rebased pointer locations to their values
        Also creates the map of dyld import addresses to the corresponding DyldBoundSymbol
    """

    @staticmethod
    def _read_chained_imports(
        binary: MachoBinary,
        chained_fixups_data_start: StaticFilePointer,
        chained_fixups_header: MachoDyldChainedFixupsHeader,
    ) -> List[DyldBoundSymbol]:
        """Parse the chained imports table in __LINKEDIT into a list of DyldBoundSymbols"""
        chained_import_addr = chained_fixups_data_start + chained_fixups_header.imports_offset
        symbols_start_addr = chained_fixups_data_start + chained_fixups_header.symbols_offset
        dyld_bound_symbols: List[DyldBoundSymbol] = []
        for i in range(chained_fixups_header.imports_count):
            chained_import = binary.read_struct(chained_import_addr, MachoDyldChainedImport)
            symbol_addr = symbols_start_addr + chained_import.name_offset
            symbol_string = binary.get_full_string_from_start_address(symbol_addr, virtual=False)
            if not symbol_string:
                raise ValueError(f"Should not happen: Failed to read a string for chained import {chained_import_addr}")
            dyld_bound_symbols.append(
                DyldBoundSymbol(
                    binary=binary,
                    address=chained_import_addr,
                    library_ordinal=chained_import.lib_ordinal,
                    name=symbol_string,
                )
            )
            chained_import_addr += chained_import.sizeof
        return dyld_bound_symbols

    @staticmethod
    def parse_chained_fixups(
        binary: MachoBinary,
    ) -> Tuple[Dict[VirtualMemoryPointer, VirtualMemoryPointer], Dict[VirtualMemoryPointer, DyldBoundSymbol]]:
        """Parses the chained fixup pointer data in __LINKEDIT
        Returns:
            Tuple[
                Dict[address containing a pointer needing to be rebased, destination assuming the stated virtual base],
                Dict[address containing a pointer that needs to be bound at load time, corresponding DyldBoundSymbol],
            ]
        """
        if not binary._dyld_chained_fixups:
            raise ValueError("This method expects the provided binary to contain chained fixup pointers")

        chained_fixups_data_start = binary._dyld_chained_fixups.dataoff
        chained_fixups_header = binary.read_struct(chained_fixups_data_start, MachoDyldChainedFixupsHeader)

        # First, read the table of bound symbols that are present anywhere within the binary
        # Bound fixup pointers will encode an index ("ordinal") into this table to state the symbol they're referring to
        dyld_bound_symbols = DyldInfoParser._read_chained_imports(
            binary, chained_fixups_data_start, chained_fixups_header
        )
        logging.debug(f"dyld chained imports table contains {len(dyld_bound_symbols)} symbols")

        # By parsing each chain of fixup pointers in the binary, we'll populate this map of addresses -> binds
        dyld_bound_addresses_to_symbols: Dict[VirtualMemoryPointer, DyldBoundSymbol] = {}

        # Next, parse the structure directly after the chained fixups header.
        # This structure gives the locations of each chain of fixup pointers within each binary segment.
        #
        # The first word of this structure provides the number of uint32_t offsets to follow
        # Each offset is added to the base address of this structure to provide the address of a
        # `struct dyld_chained_starts_in_segment`.
        chained_starts_in_image_off = chained_fixups_data_start + chained_fixups_header.starts_offset
        chained_starts_in_image = binary.read_struct(chained_starts_in_image_off, MachoDyldChainedStartsInImage)
        chained_starts_in_seg_offsets_base = chained_starts_in_image_off + chained_starts_in_image.sizeof

        # While processing rebases, we'll need to overwrite the rebase fixup pointer with the internal pointer it's
        # referring to. Since we'll be making many such writes, use a MachoBinaryWriter to do them more efficiently.
        rebases: Dict[VirtualMemoryPointer, VirtualMemoryPointer] = {}
        for segment_idx in range(chained_starts_in_image.seg_count):
            # Read entry of variable-length array of words. See comment in MachoDyldChainedStartsInImageRaw
            starts_in_seg_struct_offset = binary.read_word(
                chained_starts_in_seg_offsets_base + (segment_idx * sizeof(c_uint32)),
                virtual=False,
                word_type=c_uint32,
            )
            # Skip segments that don't contain chains
            if starts_in_seg_struct_offset == 0:
                continue

            starts_in_seg_addr = chained_starts_in_image_off + starts_in_seg_struct_offset
            chained_starts_in_seg = binary.read_struct(starts_in_seg_addr, MachoDyldChainedStartsInSegment)
            logging.debug(
                f"ChainedStartsInSegment\tsegment {segment_idx}\t"
                f"pointer_fmt {chained_starts_in_seg.pointer_format}\tpage count {chained_starts_in_seg.page_count}"
            )

            offset_in_page_start = starts_in_seg_addr + chained_starts_in_seg.sizeof
            for page_idx in range(chained_starts_in_seg.page_count):
                # Read entry of variable-length array of words. See comment in MachoDyldChainedStartsInSegmentRaw
                offset_in_page = binary.read_word(
                    offset_in_page_start + (page_idx * sizeof(c_uint16)), virtual=False, word_type=c_uint16
                )
                logging.debug(f"\tPageIdx {page_idx}, offset in page {hex(offset_in_page)}")

                chain_base = (
                    chained_starts_in_seg.segment_offset + (page_idx * chained_starts_in_seg.page_size) + offset_in_page
                )
                # Process this chain of fixup pointers
                rebases_in_chain, bound_addresses_in_chain = DyldInfoParser._process_fixup_pointer_chain(
                    binary, dyld_bound_symbols, chain_base
                )
                rebases.update(rebases_in_chain)
                dyld_bound_addresses_to_symbols.update(bound_addresses_in_chain)

        return rebases, dyld_bound_addresses_to_symbols

    @staticmethod
    def _process_fixup_pointer_chain(
        binary: MachoBinary, dyld_bound_symbols_table: List[DyldBoundSymbol], chain_base: VirtualMemoryPointer
    ) -> Tuple[Dict[VirtualMemoryPointer, VirtualMemoryPointer], Dict[VirtualMemoryPointer, DyldBoundSymbol]]:
        rebased_pointers: Dict[VirtualMemoryPointer, VirtualMemoryPointer] = {}
        dyld_bound_addresses_to_symbols: Dict[VirtualMemoryPointer, DyldBoundSymbol] = {}
        virtual_base = binary.get_virtual_base()
        # As each fixup pointer will tell us whether there are any more to follow, loop forever
        # XXX(PT): Impose an upper bound on this loop, just in case
        for _ in range(10000):
            chained_rebase_ptr = binary.read_struct(chain_base, MachoDyldChainedPtr64Rebase)
            # Rebase or bind?
            if chained_rebase_ptr.bind == 1:
                # Bind. Keep track that there is an imported symbol bind here
                chained_bind_ptr = binary.read_struct(chain_base, MachoDyldChainedPtr64Bind)
                bound_symbol = dyld_bound_symbols_table[chained_bind_ptr.ordinal]
                logging.debug(
                    f"\t\t{hex(chain_base)}: BIND\tordinal {chained_bind_ptr.ordinal}\t"
                    f"addend {chained_bind_ptr.addend}\treserved {chained_bind_ptr.reserved}\t"
                    f"next {chained_bind_ptr.next}\tsymbol {bound_symbol.name}\t\t"
                    f"dylib {binary.dylib_name_for_library_ordinal(bound_symbol.library_ordinal)}"
                )
                dyld_bound_addresses_to_symbols[chain_base + virtual_base] = bound_symbol
                chain_base += chained_bind_ptr.next * 4
            else:
                # Rebase. Overwrite the fixup pointer with the internal binary pointer it refers to
                # Rebase. Keep track that there's a rebased pointer here
                chained_ptr_raw = binary.read_word(chain_base, word_type=c_uint64, virtual=False)
                logging.debug(
                    f"\t\t{hex(chain_base)}: DyldChainedPtr64Rebase(raw: {hex(chained_ptr_raw)}) "
                    f"target={StaticFilePointer(chained_rebase_ptr.target)}"
                )
                rebased_pointers[VirtualMemoryPointer(chain_base + virtual_base)] = VirtualMemoryPointer(
                    chained_rebase_ptr.target + virtual_base
                )
                chain_base += chained_rebase_ptr.next * 4

            # Reached the end of the chain?
            if chained_rebase_ptr.next == 0:
                break
        else:
            raise ValueError("Failed to find end of fixup pointer chain")

        return rebased_pointers, dyld_bound_addresses_to_symbols

    @staticmethod
    def read_uleb(data: bytearray, offset: int) -> Tuple[int, int]:
        byte = data[offset]
        offset += 1

        result = byte & 0x7F
        shift = 7
        while byte & 0x80:
            byte = data[offset]
            result |= (byte & 0x7F) << shift
            shift += 7
            offset += 1

        # attempt to catch signed values and convert them if encountered
        if result > 0x100000000:
            result = c_long(result).value

        return result, offset

    @staticmethod
    def parse_dyld_info(binary: MachoBinary) -> Dict[VirtualMemoryPointer, DyldBoundSymbol]:
        if not binary.dyld_info:
            raise ValueError("This method expects the provided binary to contain LC_DYLD_INFO")

        return {
            **DyldInfoParser._parse_dyld_bytestream(binary, binary.dyld_info.bind_off, binary.dyld_info.bind_size),
            **DyldInfoParser._parse_dyld_bytestream(
                binary, binary.dyld_info.lazy_bind_off, binary.dyld_info.lazy_bind_size
            ),
        }

    @staticmethod
    def _parse_dyld_bytestream(
        binary: MachoBinary, file_offset: StaticFilePointer, size: int
    ) -> Dict[VirtualMemoryPointer, DyldBoundSymbol]:
        from ctypes import sizeof

        dyld_stubs_to_symbols: Dict[VirtualMemoryPointer, DyldBoundSymbol] = {}

        binding_info = binary.get_bytes(file_offset, size)
        pointer_size = sizeof(binary.platform_word_type)

        index = 0
        name_bytes: bytearray
        segment_index = 0
        segment_offset = 0
        library_ordinal = 0
        target_table_count = 0

        def commit_stub() -> None:
            segment_command = binary.segment_for_index(segment_index)
            segment_start = segment_command.vmaddr
            stub_addr = VirtualMemoryPointer(segment_start + segment_offset)
            name = name_bytes.decode("utf-8")

            symbol = DyldBoundSymbol(binary, stub_addr, library_ordinal, name)
            dyld_stubs_to_symbols[stub_addr] = symbol

        while index != len(binding_info):
            byte = binding_info[index]
            opcode = byte & 0xF0
            immediate = byte & 0x0F
            index += 1

            if opcode == BindOpcode.BIND_OPCODE_DONE:
                pass
            elif opcode == BindOpcode.BIND_OPCODE_SET_DYLIB_ORDINAL_IMM:
                library_ordinal = immediate
            elif opcode == BindOpcode.BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB:
                library_ordinal, index = DyldInfoParser.read_uleb(binding_info, index)
            elif opcode == BindOpcode.BIND_OPCODE_SET_DYLIB_SPECIAL_IMM:
                library_ordinal = -immediate
            elif opcode == BindOpcode.BIND_OPCODE_SET_SYMBOL_TRAILING_FLAGS_IMM:
                name_end = binding_info.find(b"\0", index)
                name_bytes = binding_info[index:name_end]
                index = name_end
            elif opcode == BindOpcode.BIND_OPCODE_SET_TYPE_IMM:
                pass
            elif opcode == BindOpcode.BIND_OPCODE_SET_ADDEND_SLEB:
                _, index = DyldInfoParser.read_uleb(binding_info, index)
            elif opcode == BindOpcode.BIND_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB:
                segment_index = immediate
                segment_offset, index = DyldInfoParser.read_uleb(binding_info, index)
            elif opcode == BindOpcode.BIND_OPCODE_ADD_ADDR_ULEB:
                addend, index = DyldInfoParser.read_uleb(binding_info, index)
                segment_offset += addend
            elif opcode == BindOpcode.BIND_OPCODE_DO_BIND:
                commit_stub()
                segment_offset += pointer_size
            elif opcode == BindOpcode.BIND_OPCODE_DO_BIND_ADD_ADDR_ULEB:
                commit_stub()
                segment_offset += pointer_size

                addend, index = DyldInfoParser.read_uleb(binding_info, index)
                segment_offset += addend
            elif opcode == BindOpcode.BIND_OPCODE_DO_BIND_ADD_ADDR_IMM_SCALED:
                commit_stub()
                # I think the format is <immediate>, <repeat times>
                # So, we always reserve at least one pointer, then skip the 'repeat' count pointers.
                segment_offset += pointer_size + (immediate * pointer_size)
            elif opcode == BindOpcode.BIND_OPCODE_DO_BIND_ULEB_TIMES_SKIPPING_ULEB:
                count, index = DyldInfoParser.read_uleb(binding_info, index)
                skip, index = DyldInfoParser.read_uleb(binding_info, index)
                for i in range(count):
                    commit_stub()
                    segment_offset += pointer_size + skip
            elif opcode == BindOpcode.BIND_OPCODE_THREADED:
                if immediate == BindOpcode.BIND_SUBOPCODE_THREADED_SET_BIND_ORDINAL_TABLE_SIZE_ULEB:
                    target_table_count, index = DyldInfoParser.read_uleb(binding_info, index)
                    if target_table_count >= (pow(2, 16) - 1):
                        raise ValueError("Invalid target_table_count")
                elif immediate == BindOpcode.BIND_SUBOPCODE_THREADED_APPLY:
                    # TODO(PT): Parse a fixup pointer chain here
                    pass
                else:
                    raise ValueError(f"Invalid threaded sub-opcode: {immediate}")
            else:
                logging.error(f"unknown dyld bind opcode {hex(opcode)}, immediate {hex(immediate)}")

        return dyld_stubs_to_symbols



from typing import Dict, List, Optional

from strongarm.macho.macho_binary import MachoBinary, VirtualMemoryPointer
from strongarm.macho.macho_definitions import NLIST_NTYPE, NTYPE_VALUES


class MachoStringTableEntry:
    """Class encapsulating an entry into the Mach-O string table
    """

    __slots__ = ["start_idx", "length", "full_string"]

    def __init__(self, start_idx: int, length: int, content: str) -> None:
        self.start_idx = start_idx
        self.length = length
        self.full_string = content


class MachoStringTableHelper:
    """Class containing helper functions for processing different tables in a Mach-O
    """

    # TODO(PT): generalize the preprocessing of a string table where we efficiently map string start addresses to
    # full strings, so we don't need to do an O(n) search for a (struct __objc_data).name or something

    def __init__(self, binary: MachoBinary) -> None:
        self.binary = binary
        self.string_table_entries = MachoStringTableHelper.transform_string_section(self.binary.get_raw_string_table())
        self.imported_symbols: List[str] = []
        self.exported_symbols: Dict[VirtualMemoryPointer, str] = {}
        self.parse_sym_lists()

    @classmethod
    def transform_string_section(cls, strtab: List[int]) -> Dict[int, MachoStringTableEntry]:
        """Create more efficient representation of string table data

        Often, tables in a Mach-O will reference data within the string table.
        The string table is a large array of characters, representing NULL-terminated strings. There is no seperator
        between entries aside from a NULL terminator. When other sections reference a string table entry, they will
        only reference the starting index. Thus, if we did no other processing, every time we got an index we'd need to
        do an O(n) loop to find the next NULL character, indicating the end of the string.

        To avoid this, we preprocess the string table into the full strings it represents. To make these lookups easier,
        we create a map of start indexes to MachoStringTableEntry's

        Returns:
            Map of string table entry start indexes to MachoStringTableEntry's
        """
        string_table_entries = {}
        entry_start_idx = 0
        for idx, ch in enumerate(strtab):
            # end of current string?
            if ch == 0x00:
                length = idx - entry_start_idx

                # read this string now that we know the start index and length
                entry_end_idx = entry_start_idx + length
                entry_byte_content = bytearray(strtab[entry_start_idx:entry_end_idx:])
                try:
                    entry_content = entry_byte_content.decode("utf-8")
                except UnicodeDecodeError:
                    # get a string literal of the raw bytes. 0x0080 -> "b'\\x00\\x80'"
                    entry_content = str(bytes(entry_byte_content))

                # record in list
                ent = MachoStringTableEntry(entry_start_idx, length, entry_content)
                string_table_entries[entry_start_idx] = ent

                # move to starting index of next string
                entry_start_idx = idx + 1
        return string_table_entries

    def string_table_entry_for_strtab_index(self, start_idx: int) -> Optional[MachoStringTableEntry]:
        """For a index in the packed character table, get the corresponding MachoStringTableEntry

        Returns:
            A MachoStringTableEntry if provided index was the starting character of a string table entry, None if not
        """
        if start_idx in self.string_table_entries:
            return self.string_table_entries[start_idx]
        return None

    def parse_sym_lists(self) -> None:
        """Read imported and exported symbol names referenced by symtab from the string table.
        """

        self.imported_symbols = []

        symtab = self.binary.symtab_contents
        for idx, sym in enumerate(symtab):
            strtab_idx = sym.n_un.n_strx
            string_table_entry = self.string_table_entry_for_strtab_index(strtab_idx)
            if not string_table_entry:
                continue
            symbol_str = string_table_entry.full_string

            is_shared_symbol = int(sym.n_type & NLIST_NTYPE.N_EXT)
            symbol_type = sym.n_type & NLIST_NTYPE.N_TYPE

            if symbol_type == NTYPE_VALUES.N_UNDF:
                # symbols marked (imported, shared) are actually duplicated as exported symbols later in the symbol
                # table. I don't know why this is.
                # Let's just ignore (imported, shared) symbols. If they do contain duplicate entries as exported
                # symbols later, they'll be caught then.
                if not is_shared_symbol:
                    continue
                self.imported_symbols.append(symbol_str)
            elif symbol_type == NTYPE_VALUES.N_SECT:
                self.exported_symbols[sym.n_value] = symbol_str

    def get_symbol_name_for_address(self, address: VirtualMemoryPointer) -> Optional[str]:
        """ For an address of a function entrypoint, return the function's symbol name
        """
        if address in self.exported_symbols:
            return self.exported_symbols[address]
        return None



from ctypes import c_uint32, c_uint64, sizeof
from dataclasses import dataclass
from types import TracebackType
from typing import List, Optional, Type, Union

from .macho_binary import MachoBinary, StaticFilePointer, VirtualMemoryPointer


@dataclass
class MachoBinaryQueuedWrite:
    file_offset: StaticFilePointer
    bytes_to_write: bytearray


class MachoBinaryWriter:
    def __init__(self, binary: MachoBinary) -> None:
        self.binary = binary
        self.modified_binary = binary
        self.queued_writes: List[MachoBinaryQueuedWrite] = []

    def __enter__(self) -> None:
        pass

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        # Create a new binary with the overwritten data
        new_binary_data = bytearray(len(self.binary._cached_binary))
        new_binary_data[:] = self.binary._cached_binary
        for write in self.queued_writes:
            new_binary_data[write.file_offset : write.file_offset + len(write.bytes_to_write)] = write.bytes_to_write

        self.modified_binary = MachoBinary(self.binary.path, new_binary_data,)

    def write_word(self, word: Union[c_uint32, c_uint64], address: int, virtual: bool = True) -> None:
        """Enqueue a write of the provided word to the binary.
        Note: This will invalidate the binary's code signature, if present.
        """
        # Ensure there is valid data in this address region by trying to read from it
        self.binary.get_contents_from_address(address, sizeof(word), virtual)

        # If the above did not throw an exception, the provided address range is valid.
        file_offset = StaticFilePointer(address)
        if virtual:
            file_offset = self.binary.file_offset_for_virtual_address(VirtualMemoryPointer(address))

        # Queue a binary write here
        self.queued_writes.append(
            MachoBinaryQueuedWrite(
                file_offset=file_offset,
                bytes_to_write=bytearray(word.value.to_bytes(length=sizeof(word), byteorder="little")),
            )
        )



from typing import TYPE_CHECKING, Optional, Tuple, Union

from capstone import CsInsn
from capstone.arm64 import ARM64_OP_IMM, ARM64_OP_MEM, ARM64_OP_REG, Arm64Op

from strongarm.macho.macho_analyzer import MachoAnalyzer
from strongarm.macho.macho_definitions import VirtualMemoryPointer
from strongarm.macho.objc_runtime_data_parser import ObjcSelector, ObjcSelref

if TYPE_CHECKING:
    from .objc_analyzer import ObjcFunctionAnalyzer


class ObjcInstruction:
    VECTOR_REGISTER_PREFIXES = ["d", "s", "v"]

    def __init__(self, instruction: CsInsn) -> None:
        self.raw_instr = instruction
        self.address = VirtualMemoryPointer(self.raw_instr.address)

        self.is_msgSend_call: bool = False
        self.symbol: Optional[str] = None

    def __repr__(self) -> str:
        return f"<ObjcInstruction {self.symbol} at 0x{self.address:x}>"

    @classmethod
    def is_vector_register(cls, reg_name: str) -> bool:
        """Returns True if the register refers to a vector register; False otherwise.
        """
        for vector_prefix in ObjcInstruction.VECTOR_REGISTER_PREFIXES:
            if vector_prefix in reg_name:
                return True
        return False

    @classmethod
    def _operand_uses_vector_registers(cls, instruction: CsInsn, operand: Arm64Op) -> bool:
        if operand.type == ARM64_OP_IMM:
            return False

        if operand.type == ARM64_OP_REG:
            reg_name = instruction.reg_name(operand.value.reg)
        elif operand.type == ARM64_OP_MEM:
            reg_name = instruction.reg_name(operand.mem.base)
        else:
            raise RuntimeError(f"unknown operand type {operand.type} in instr at {instruction.address}")
        return ObjcInstruction.is_vector_register(reg_name)

    @classmethod
    def instruction_uses_vector_registers(cls, instruction: CsInsn) -> bool:
        """Returns True if the instruction accesses vector registers.
        False if the instruction only uses general-purpose registers.
        """
        for op in instruction.operands:
            if ObjcInstruction._operand_uses_vector_registers(instruction, op):
                return True
        return False

    @classmethod
    def parse_instruction(
        cls, function_analyzer: "ObjcFunctionAnalyzer", instruction: CsInsn
    ) -> Union["ObjcInstruction", "ObjcUnconditionalBranchInstruction", "ObjcConditionalBranchInstruction"]:
        """Read an instruction and encapsulate it in the appropriate ObjcInstruction subclass
        """
        if ObjcBranchInstruction.is_branch_instruction(instruction):
            return ObjcBranchInstruction.parse_instruction(function_analyzer, instruction)
        return ObjcInstruction(instruction)


class ObjcBranchInstruction(ObjcInstruction):
    def __init__(self, instruction: CsInsn, destination_address: VirtualMemoryPointer) -> None:
        super(ObjcBranchInstruction, self).__init__(instruction)

        self.destination_address = destination_address

        self.selref: Optional[ObjcSelref] = None
        self.selector: Optional[ObjcSelector] = None

        self.is_external_c_call: bool = False
        self.is_external_objc_call: bool = False

    @classmethod
    def parse_instruction(
        cls,
        function_analyzer: "ObjcFunctionAnalyzer",
        instruction: CsInsn,
        patch_msgSend_destination: bool = True,
        container_function_boundary: Tuple[VirtualMemoryPointer, VirtualMemoryPointer] = None,
    ) -> Union["ObjcUnconditionalBranchInstruction", "ObjcConditionalBranchInstruction"]:
        """Read a branch instruction and encapsulate it in the appropriate ObjcBranchInstruction subclass
        """
        # use appropriate subclass
        if instruction.mnemonic in ObjcUnconditionalBranchInstruction.UNCONDITIONAL_BRANCH_MNEMONICS:
            uncond_instr = ObjcUnconditionalBranchInstruction(
                function_analyzer, instruction, patch_msgSend_destination, container_function_boundary
            )
            return uncond_instr

        elif instruction.mnemonic in ObjcConditionalBranchInstruction.CONDITIONAL_BRANCH_MNEMONICS:
            cond_instr = ObjcConditionalBranchInstruction(function_analyzer, instruction)
            return cond_instr

        else:
            raise ValueError(f"Unknown branch mnemonic {instruction.mnemonic}")

    @classmethod
    def is_branch_instruction(cls, instruction: CsInsn) -> bool:
        """Returns True if the CsInsn represents a branch instruction, False otherwise
        """
        # TODO(FS): Merge subclasses into ObjcBranchInstruction and provide contextual information about each variant
        return (
            instruction.mnemonic in ObjcUnconditionalBranchInstruction.UNCONDITIONAL_BRANCH_MNEMONICS
            or instruction.mnemonic in ObjcConditionalBranchInstruction.CONDITIONAL_BRANCH_MNEMONICS
        )


class ObjcUnconditionalBranchInstruction(ObjcBranchInstruction):
    UNCONDITIONAL_BRANCH_MNEMONICS = [
        "b",
        "bl",
        "bx",
        "blx",
        "bxj",
        "b.eq",  # TODO(PT): these b-suffix are not strictly unconditional branches, but
        # they're functionally unconditional for what we care about
        "b.ne",
        "b.ge",
        "b.le",
        "b.gt",
        "b.lt",
        "b.hi",
        "b.lo",
    ]
    OBJC_MSGSEND_FUNCTIONS = ["_objc_msgSend", "_objc_msgSendSuper2"]

    def __init__(
        self,
        function_analyzer: "ObjcFunctionAnalyzer",
        instruction: CsInsn,
        patch_msgSend_destination: bool = True,
        container_function_boundary: Tuple[VirtualMemoryPointer, VirtualMemoryPointer] = None,
    ) -> None:
        if instruction.mnemonic not in ObjcUnconditionalBranchInstruction.UNCONDITIONAL_BRANCH_MNEMONICS:
            raise ValueError(
                f"ObjcUnconditionalBranchInstruction instantiated with" f" invalid mnemonic {instruction.mnemonic}"
            )
        # an unconditional branch has the destination as the only operand
        super().__init__(instruction, VirtualMemoryPointer(instruction.operands[0].value.imm))

        self.selref: Optional[ObjcSelref] = None
        self.selector: Optional[ObjcSelector] = None

        analyzer = MachoAnalyzer.get_analyzer(function_analyzer.binary)

        if container_function_boundary:
            if self.destination_address >= container_function_boundary[0]:
                if self.destination_address < container_function_boundary[1]:
                    # Local basic-block branch within a function
                    # print(f'{self.destination_address} local branch fast path')
                    self.symbol = None
                    self.is_external_c_call = False
                    self.is_msgSend_call = False
                    return

        called_sym = analyzer.callable_symbol_for_address(self.destination_address)
        if not called_sym:
            # Branch to an anonymous destination
            # Might be a basic block within a function or some other label
            # logging.debug(f'No symbol for branch destination {hex(self.destination_address)}')
            self.is_external_c_call = False
            self.is_msgSend_call = False
            self.symbol = None
            return

        self.symbol = called_sym.symbol_name
        self.is_external_c_call = called_sym.is_imported

        if called_sym.is_imported:
            if called_sym.symbol_name in self.OBJC_MSGSEND_FUNCTIONS:
                self.is_msgSend_call = True
                self.is_external_c_call = False
                if patch_msgSend_destination:
                    self._patch_msgSend_destination(function_analyzer)
            else:
                self.is_msgSend_call = False
        else:
            self.is_msgSend_call = False

    def _patch_msgSend_destination(self, function_analyzer: "ObjcFunctionAnalyzer") -> None:
        # validate instruction
        if (
            not self.is_msgSend_call
            or self.raw_instr.mnemonic not in ["bl", "b"]
            or self.symbol not in self.OBJC_MSGSEND_FUNCTIONS
        ):
            raise ValueError(
                f"cannot parse objc_msgSend destination on non-msgSend instruction"
                f" {function_analyzer.format_instruction(self.raw_instr)}"
            )
        # if this is an objc_msgSend target, patch destination_address to be the address of the targeted IMP
        # note! this means destination_address is *not* the actual destination address of the instruction
        # the *real* destination will be a stub function corresponding to _objc_msgSend, but
        # knowledge of this is largely useless, and the much more valuable piece of information is
        # which function the selector passed to objc_msgSend corresponds to.
        # therefore, replace the 'real' destination address with the requested IMP
        try:
            selref_ptr = function_analyzer.get_objc_selref(self)
            selector = function_analyzer.macho_analyzer.selector_for_selref(selref_ptr)
            if not selector:
                raise RuntimeError(f"Couldn't get sel for selref ptr {selref_ptr}")
            # if we couldn't find an IMP for this selref,
            # it is defined in a class outside this binary
            self.is_external_objc_call = selector.is_external_definition

            # Only patch destination_address if the implementation is in this binary.
            # Otherwise, destination_address will continue to point to __imp_stubs_objc_msgSend
            if selector.implementation:
                self.destination_address = selector.implementation
            self.selref = selector.selref
            self.selector = selector
        except RuntimeError:
            # TODO(PT): Should this ever be hit?
            self.is_external_objc_call = True
            self.destination_address = VirtualMemoryPointer(0)


class ObjcConditionalBranchInstruction(ObjcBranchInstruction):
    SINGLE_OP_MNEMONICS = ["cbz", "cbnz"]
    DOUBLE_OP_MNEMONICS = ["tbnz"]
    CONDITIONAL_BRANCH_MNEMONICS = SINGLE_OP_MNEMONICS + DOUBLE_OP_MNEMONICS

    def __init__(self, function_analyzer: "ObjcFunctionAnalyzer", instruction: CsInsn) -> None:
        if instruction.mnemonic not in ObjcConditionalBranchInstruction.CONDITIONAL_BRANCH_MNEMONICS:
            raise ValueError(
                f"ObjcConditionalBranchInstruction instantiated with" f" invalid mnemonic {instruction.mnemonic}"
            )

        # a conditional branch will either hold the destination in first or second operand, depending on mnemonic
        if instruction.mnemonic in ObjcConditionalBranchInstruction.SINGLE_OP_MNEMONICS:
            dest_op_idx = 1
        elif instruction.mnemonic in ObjcConditionalBranchInstruction.DOUBLE_OP_MNEMONICS:
            dest_op_idx = 2
        else:
            raise ValueError(f"Unknown conditional mnemonic {instruction.mnemonic}")

        ObjcBranchInstruction.__init__(
            self, instruction, VirtualMemoryPointer(instruction.operands[dest_op_idx].value.imm)
        )



# PT: Importing from strongarm_dataflow implicitly links against the capstone shared library,
# and if capstone is not installed correctly it will raise an ImportError.
# Report this in a clearer way so the user can see exactly what went wrong.
try:
    # Re-export symbols from from the sequestered strongarm_dataflow module
    from strongarm_dataflow.dataflow import get_register_contents_at_instruction_fast
    from strongarm_dataflow.register_contents import RegisterContents, RegisterContentsType
except ImportError as e:
    if "libcapstone" in str(e):
        import sys

        print("\ncapstone 4.x could not be found, is the capstone backend installed?\n")
        sys.exit(1)
    raise

from .objc_analyzer import BasicBlock, ObjcFunctionAnalyzer, ObjcMethodInfo
from .objc_instruction import (
    ObjcBranchInstruction,
    ObjcConditionalBranchInstruction,
    ObjcInstruction,
    ObjcUnconditionalBranchInstruction,
)

__all__ = [
    "get_register_contents_at_instruction_fast",
    "BasicBlock",
    "ObjcFunctionAnalyzer",
    "ObjcMethodInfo",
    "RegisterContents",
    "RegisterContentsType",
    "ObjcBranchInstruction",
    "ObjcConditionalBranchInstruction",
    "ObjcInstruction",
    "ObjcUnconditionalBranchInstruction",
]



from typing import List

from capstone import Cs, CsInsn

from strongarm.macho.macho_binary import MachoBinary
from strongarm.macho.macho_definitions import VirtualMemoryPointer


class MachoImpStub:
    """Encapsulates entry in __imp_stubs section

    An 'entry' in the __imp_stubs section is a very short function which jumps to a pointer in the __got or
    __la_symbol_ptr lists. This pointer is a 'garbage' value which will be filled by dyld at runtime the first time
    the stub is invoked by a function called dyld_stub_binder.
    An entry in the __imp_stubs section might be assembly like the following:
    0x0000000100006898         nop
    0x000000010000689c         ldr        x16, #0x100008010
    0x00000001000068a0         br         x16
    In this case, 0x100008010 is an address in __la_symbol_ptr, which contains `dq 0x100010000`. I don't know the exact
    mechanism by which dyld_stub_binder rewrites __imp_stubs/__la_symbol_ptr at runtime to change garbage pointers such
    as 0x100010000 into the address the symbol was loaded at, but it's not really relevant here.
    More relevant is the fact that the first address of each stub entry (0x100006898 in this example) will be the
    branch destination anytime someone addresses the external symbol in question.
    So, if the imp stub above corresponded to the `__la_symbol_ptr` entry for NSLog, a caller calling NSLog would
    actually branch to 0x100006898.
    By chaining all this information together along with symbol names cross-referenced with __la_symbol_ptr from
    the indirect symbol table + string table, we can cross-reference branch destinations to external symbol names.

    This object contains the starting address of the stub (which will be the destination for branches),
    as well as the __la_symbol_ptr entry which is targeted by the stub.
    """

    def __init__(self, address: VirtualMemoryPointer, destination: VirtualMemoryPointer) -> None:
        self.address = address
        self.destination = destination


class MachoImpStubsParser:
    def __init__(self, binary: MachoBinary, capstone_disasm: Cs) -> None:
        self.binary = binary
        self._cs = capstone_disasm
        self.imp_stubs = self._parse_all_stubs()

    @staticmethod
    def _parse_stub_from_instructions(instr1: CsInsn, instr2: CsInsn, instr3: CsInsn) -> MachoImpStub:
        # TODO(PT): write CsInsn by hand to test this function
        # each stub follows one of two patterns
        # pattern 1: nop / ldr x16, <sym> / br x16
        # pattern 2: adrp x16, <page> / ldr x16, [x16 <offset>] / br x16
        # try parsing both of these formats
        patterns = [["nop", "ldr", "br"], ["adrp", "ldr", "br"]]
        # differentiate between patterns by looking at the opcode of the first instruction
        if instr1.mnemonic == patterns[0][0]:
            pattern_idx = 0
        elif instr1.mnemonic == patterns[1][0]:
            pattern_idx = 1
        else:
            # unknown stub format
            raise NotImplementedError({hex(instr1.address)})

        expected_ops = patterns[pattern_idx]
        for idx, op in enumerate([instr1, instr2, instr3]):
            # sanity check
            if op.mnemonic != expected_ops[idx]:
                raise RuntimeError(
                    f"Expected instr {hex(op.address)} (idx {idx}) to be {expected_ops[idx]}"
                    f" while parsing stub, was instead {op.mnemonic}"
                )

        stub_addr = instr1.address
        stub_dest = 0
        # nop/ldr/br pattern
        if pattern_idx == 0:
            stub_dest = instr2.operands[1].value.imm
        # adrp/ldr/br pattern
        elif pattern_idx == 1:
            stub_dest_page = instr1.operands[1].value.imm
            stub_dest_pageoff = instr2.operands[1].mem.disp
            stub_dest = stub_dest_page + stub_dest_pageoff
        stub = MachoImpStub(VirtualMemoryPointer(stub_addr), VirtualMemoryPointer(stub_dest))
        return stub

    def _parse_all_stubs(self) -> List[MachoImpStub]:
        stubs_section = self.binary.section_with_name("__stubs", "__TEXT")
        if not stubs_section:
            return []

        func_str = bytes(
            self.binary.get_bytes(
                stubs_section.offset, stubs_section.cmd.size, _translate_addr_to_file=False
            )  # When working with DSC's, the reported offset should not be translated
        )
        instructions = [instr for instr in self._cs.disasm(func_str, stubs_section.address)]

        stubs = []
        # each stub follows one of two patterns
        # pattern 1: nop / ldr x16, <sym> / br x16
        # pattern 2: adrp x16, <page> / ldr x16, [x16 <offset>] / br x16
        # try parsing both of these formats

        irpd = iter(instructions)
        for instr1, instr2, instr3 in zip(irpd, irpd, irpd):
            stub = self._parse_stub_from_instructions(instr1, instr2, instr3)
            if not stub:
                raise RuntimeError("Failed to parse stub")
            stubs.append(stub)
        return stubs



import logging
from typing import Optional

from strongarm.macho.macho_binary import MachoBinary
from strongarm.macho.macho_definitions import StaticFilePointer

from .codesign_definitions import CodesignBlobTypeEnum, CSBlob, CSBlobIndex, CSCodeDirectory, CSSuperblob


class CodesignParser:
    """Parser for the CodeSign blobs in __LINKEDIT pointed to by LC_CODE_SIGNATURE.
    https://opensource.apple.com/source/xnu/xnu-2422.1.72/bsd/sys/codesign.h
    https://opensource.apple.com/source/xnu/xnu-4570.1.46/osfmk/kern/cs_blobs.h.auto.html
    https://opensource.apple.com/source/libsecurity_utilities/libsecurity_utilities-55030/lib/blob.h.auto.html
    https://opensource.apple.com/source/Security/Security-57031.1.35/Security/libsecurity_codesigning/lib/CSCommonPriv.h
    """

    def __init__(self, binary: MachoBinary) -> None:
        self.binary = binary
        self.entitlements: bytearray = bytearray(b"<plist></plist>")
        self.signing_identifier: Optional[str] = None
        self.signing_team_id: Optional[str] = None

        # If the binary does not have a code signature, we have nothing to do
        if not self.binary.code_signature_cmd:
            return

        self._codesign_entry = self.binary.code_signature_cmd.dataoff
        self.parse_codesign_blob(self._codesign_entry)

    def read_32_big_endian(self, offset: StaticFilePointer) -> int:
        """Read a 32-bit word from the file offset in big-endian order.
        """
        word_bytes = self.binary.get_bytes(offset, 4)
        word = int.from_bytes(word_bytes, byteorder="big")
        return word

    def parse_codesign_blob(self, file_offset: StaticFilePointer) -> None:
        """High-level parser to parse the codesign blob at the file offset.
        """
        magic = self.read_32_big_endian(file_offset)

        if magic == CodesignBlobTypeEnum.CSMAGIC_CODE_DIRECTORY:
            self.parse_code_directory(file_offset)
        elif magic == CodesignBlobTypeEnum.CSMAGIC_EMBEDDED_SIGNATURE:
            self.parse_superblob(file_offset)
        elif magic == CodesignBlobTypeEnum.CSMAGIC_EMBEDDED_ENTITLEMENTS:
            self.entitlements = self.parse_entitlements(file_offset)
        elif magic == CodesignBlobTypeEnum.CSMAGIC_REQUIREMENT:
            pass
        elif magic == CodesignBlobTypeEnum.CSMAGIC_REQUIREMENT_SET:
            pass
        elif magic == CodesignBlobTypeEnum.CSMAGIC_DETACHED_SIGNATURE:
            pass
        elif magic == CodesignBlobTypeEnum.CSMAGIC_BLOBWRAPPER:
            pass
        else:
            # unknown magic
            logging.debug(self, f"Unknown CodeSign blob magic @ {hex(file_offset)}: {hex(magic)}")

    def parse_superblob(self, file_offset: StaticFilePointer) -> None:
        """Parse a codesign 'superblob' at the provided file offset.
        This is a blob which embeds several child blobs.
        The superblob format is the superblob header, followed by several csblob_index structures describing
        the layout of the child blobs.
        """
        internal_file_offset = int(file_offset)
        superblob = self.binary.read_struct(internal_file_offset, CSSuperblob)
        if superblob.magic != CodesignBlobTypeEnum.CSMAGIC_EMBEDDED_SIGNATURE:
            raise RuntimeError(f"Can blobs other than embedded signatures be superblobs? {hex(superblob.magic)}")

        # move past the superblob header to the first index struct entry
        internal_file_offset += superblob.sizeof

        # parse each struct csblob_index following the superblob header
        for i in range(superblob.index_entry_count):
            csblob_index = self.parse_csblob_index(StaticFilePointer(internal_file_offset))
            csblob_file_offset = self._codesign_entry + csblob_index.offset

            # found a blob, now parse it
            self.parse_codesign_blob(StaticFilePointer(csblob_file_offset))

            # iterate to the next blob index struct in list
            internal_file_offset += csblob_index.sizeof

    @staticmethod
    def get_index_blob_name(blob_index: CSBlobIndex) -> str:
        """Get the human-readable blob type from the `type` field in a CSBlobIndex.
        """
        # cs_blobs.h
        blob_types = {
            0: "Code Directory",
            1: "Info slot",
            2: "Requirement Set",
            3: "Resource Directory",
            4: "Application",
            5: "Embedded Entitlements",
            0x1000: "Alternate Code Directory",
            0x10000: "CMS Signature",
        }
        return blob_types[blob_index.type]

    def parse_csblob_index(self, file_offset: StaticFilePointer) -> CSBlobIndex:
        """Parse a csblob_index at the file offset.
        A csblob_index is a header structure describing the type/layout of a superblob's child blob.
        This method will parse and return the index header.
        """
        blob_index = self.binary.read_struct(file_offset, CSBlobIndex)
        return blob_index

    def parse_code_directory(self, file_offset: StaticFilePointer) -> None:
        """Parse a Code Directory at the file offset.
        """
        code_directory = self.binary.read_struct(file_offset, CSCodeDirectory)

        identifier_address = code_directory.binary_offset + code_directory.identifier_offset
        identifier_string = self.binary.get_full_string_from_start_address(identifier_address, virtual=False)
        self.signing_identifier = identifier_string

        # Version 0x20100+ includes scatter_offset
        # Version 0x20200+ includes team offset
        if code_directory.version >= 0x20200:
            # Note that if the version < 0x20200, the CSCodeDirectory structure parses past the end of the actual struct
            team_id_address = code_directory.binary_offset + code_directory.team_offset
            team_id_string = self.binary.get_full_string_from_start_address(team_id_address, virtual=False)
            self.signing_team_id = team_id_string

    def print_code_directory(self, code_dir: CSCodeDirectory) -> None:
        print(f"CodeDirectory @ {hex(code_dir.binary_offset)}")
        print("-----------------------")
        print(f"Version: {hex(code_dir.version)}")
        print(f"Flags: {hex(code_dir.flags)}")
        print(f"Hash offset: {hex(code_dir.hash_offset)}")
        print(f"Identifier offset: {hex(code_dir.identifier_offset)}")
        print(f"Special slots count: {code_dir.special_slots_count}")
        print(f"Code limit: {hex(code_dir.code_limit)}")
        print(f"Hash size: {hex(code_dir.hash_size)}")
        print(f"Hash type: {hex(code_dir.hash_type)}")
        print(f"Platform: {hex(code_dir.platform)}")
        print(f"Page size: {hex(code_dir.page_size)}")
        print(f"Scatter offset: {hex(code_dir.scatter_offset)}")
        print(f"Team offset: {hex(code_dir.team_offset)}")
        print()

    def parse_entitlements(self, file_offset: StaticFilePointer) -> bytearray:
        """Parse the embedded entitlements blob at the file offset.
        Returns a bytearray of the embedded entitlements.
        """
        entitlements_blob = self.binary.read_struct(file_offset, CSBlob)
        if entitlements_blob.magic != CodesignBlobTypeEnum.CSMAGIC_EMBEDDED_ENTITLEMENTS:
            raise RuntimeError(f"incorrect magic for embedded entitlements: {hex(entitlements_blob.magic)}")

        blob_end = entitlements_blob.binary_offset + entitlements_blob.length

        xml_start = StaticFilePointer(file_offset + entitlements_blob.sizeof)
        xml_length = blob_end - xml_start
        xml = self.binary.get_bytes(xml_start, xml_length)
        return xml
